import os

from dotenv import load_dotenv


def get_scratch_dir():
    """Get scratch directory from environment variable `DEEPSEARCH_GLM_SCRATCH_DIR` (defined in .env)"""

    load_dotenv()

    tmpdir = os.path.abspath(os.getenv("DEEPSEARCH_GLM_SCRATCH_DIR"))

    if not os.path.exists(tmpdir):
        os.mkdir(tmpdir)

    return tmpdir



#!/usr/bin/env python
"""Module to create a Tokenizer model (train)"""

import argparse
import json
import os
import random

import tqdm
from tabulate import tabulate

from deepsearch_glm.nlp_utils import train_tok


def parse_arguments():
    """Function to parse arguments for `nlp_train_tok`"""

    parser = argparse.ArgumentParser(
        prog="nlp_train_tok",
        description="train Tokenizer (SentencePiece) model",
        epilog="""
examples of execution: 

1. end-to-end example to train CRF:

    poetry run python ./deepsearch_glm/nlp_train_tok.py -t unigram -n <name> --input-file <filename>
""",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.add_argument(
        "-t",
        "--type",
        required=True,
        default="unigram",
        help="model-type for training tokenizer model",
        choices=["unigram", "bpe", "word", "char"],
    )

    parser.add_argument(
        "-n",
        "--name",
        required=True,
        type=str,
        default="./trained-tokenizer",
        help="name of the output-model",
    )

    parser.add_argument(
        "-i",
        "--input-file",
        required=True,
        type=str,
        default=None,
        help="input-file with text (1 sentence per line)",
    )

    args = parser.parse_args()

    return args.model_type, args.model_name, args.input_file


def create_tok_model(model_type: str, model_name: str, ifile: str):
    """Create a new tokenizer from a text-file"""

    return train_tok(model_type, model_name, ifile)


if __name__ == "__main__":
    mtype, mname, ifile = parse_arguments()

    create_tok_model(mtype, mname, ifile)



#!/usr/bin/env python

import os

import sys

import subprocess
from typing import List
import pybind11

ROOT_DIR = os.path.abspath("./")
BUILD_DIR = os.path.join(ROOT_DIR, "build")

def get_pybind11_cmake_args():
        pybind11_sys_path = os.getenv("PYBIND11_SYSPATH")
        if pybind11_sys_path:
            # pybind11_include_dir = os.path.join(pybind11_sys_path, "include")
            pybind11_cmake_dir = os.path.join(pybind11_sys_path, "share", "cmake", "pybind11")
        else:
            # pybind11_include_dir = pybind11.get_include()
            pybind11_cmake_dir = pybind11.get_cmake_dir()
        # print(f"{pybind11_include_dir=}")
        print(f"{pybind11_cmake_dir=}")
        return [f"-Dpybind11_DIR={pybind11_cmake_dir}"]

def run(cmd: List[str], cwd: str="./"):

    print_cmd = " ".join(cmd)
    print(f"\nlaunch: {print_cmd}")

    message = subprocess.run(cmd, cwd=cwd)

    if "returncode=0" in str(message):
        print(f" -> SUCCESS")
        return True

    print(f" -> ERROR with message: '{message}'\n")
    return False


def build_local(num_threads: int):

    USE_SYSTEM_DEPS = os.getenv("USE_SYSTEM_DEPS", "OFF")


    print("python prefix: ", sys.exec_prefix)
    print("python executable: ", sys.executable)
    config_cmd = [
        "cmake",
        "-B", f"{BUILD_DIR}",
        f"-DUSE_SYSTEM_DEPS={USE_SYSTEM_DEPS}",
        f"-DPYTHON_EXECUTABLE={sys.executable}",
    ]
    config_cmd.extend(get_pybind11_cmake_args())
    success = run(config_cmd, cwd=ROOT_DIR)
    if not success:
        raise RuntimeError("Error building.")

    build_cmd = [
        "cmake",
        "--build", f"{BUILD_DIR}",
        "--target=install",
    ]
    if num_threads > 1:
        build_cmd.extend(["-j", f"{num_threads}"])
    success = run(build_cmd, cwd=ROOT_DIR)
    if not success:
        raise RuntimeError("Error building.")


if "__main__" == __name__:
    num_threads = int(os.getenv("BUILD_THREADS", "4"))
    build_local(num_threads=num_threads)



"""Module to provide basic functionality of the Deep Search platform"""

import copy
import datetime
import glob
import hashlib
import json
import os
import subprocess

from deepsearch_glm.utils.common import get_scratch_dir

try:
    import deepsearch as ds
except ImportError as err:
    raise ImportError(
        "To use the Deep Search capabilities, it is required to install the deepsearch-toolkit package."
        "You can install it with `pip install deepsearch-toolkit` or using the convenience extra in "
        "deepsearch-glm as `pip install deepsearch-glm[toolkit]`."
    ) from None

from deepsearch.cps.client.components.elastic import ElasticDataCollectionSource

# from deepsearch.cps.client.components.queries import RunQueryError
from deepsearch.cps.queries import DataQuery
from dotenv import load_dotenv
from tqdm import tqdm

# re-import from the new module, in order to keep backwards compatibility
from deepsearch_glm.utils.doc_utils import (
    resolve_item,
    to_legacy_document_format,
    to_xml_format,
)


def load_vars():
    """Load variables `DEEPSEARCH_HOST`, `DEEPSEARCH_PROJ`, `DEEPSEARCH_USERNAME`, `DEEPSEARCH_APIKEY` and `DEEPSEARCH_VERIFYSSL` (defined in .env)"""

    load_dotenv()

    host = os.getenv("DEEPSEARCH_HOST")
    proj = os.getenv("DEEPSEARCH_PROJ")

    username = os.getenv("DEEPSEARCH_USERNAME")
    apikey = os.getenv("DEEPSEARCH_APIKEY")

    verify_ssl = os.getenv("DEEPSEARCH_VERIFYSSL")

    # avoid common mistakes ...
    if host == "https://deepsearch-experience.res.ibm.com/":
        verify_ssl = True
    elif host == "https://cps.foc-deepsearch.zurich.ibm.com/":
        verify_ssl = False

    return host, proj, username, apikey, verify_ssl


def get_ds_api():
    """Obtain Deep Search API object"""

    tdir = get_scratch_dir()

    host, proj, username, apikey, verify_ssl = load_vars()

    config_ = {
        "host": host,
        "auth": {"username": username, "api_key": apikey},
        "verify_ssl": verify_ssl,
    }

    config_file = f"{tdir}/ds_config.json"
    with open(config_file, "w", encoding="utf-8") as fw:
        fw.write(json.dumps(config_))

    config = ds.DeepSearchConfig.parse_file(config_file)
    client = ds.CpsApiClient(config)

    api = ds.CpsApi(client)

    return api, proj


def process_zip_files(tdir):
    """Unzip all files obtained from Deep Search"""

    zipfiles = sorted(glob.glob(os.path.join(tdir, "*.zip")))
    # print(f"zips: ", len(zipfiles))

    for zipfile in zipfiles:
        cmd = ["unzip", zipfile, "-d", tdir]
        print(" ".join(cmd))

        subprocess.call(cmd)

    # clean up
    for i, zipfile in enumerate(zipfiles):
        print(i, "\t removing ", zipfile)
        subprocess.call(["rm", zipfile])

    """
    cellsfiles = sorted(glob.glob(os.path.join(tdir, "*.cells")))
    for i,cellsfile in enumerate(cellsfiles):
        subprocess.call(["rm", cellsfile])            
    """


def convert_pdffiles(pdf_files, force=False):
    """Convert PDF files to JSON"""

    for i, pdf_file in enumerate(pdf_files):
        old_file = pdf_file

        subprocess.call(["ls", f"{old_file}"])

        if " " in pdf_file:
            new_file = pdf_file.replace(" ", "_")
            subprocess.call(["mv", f"{old_file}", f"{new_file}"])
            pdf_files[i] = new_file

    json_files = []

    new_pdfs = []
    for pdf_file in pdf_files:
        if force:
            new_pdfs.append(pdf_file)
        else:
            json_file = pdf_file.replace(".pdf", ".json")
            if os.path.exists(json_file):
                json_files.append(json_file)
            else:
                new_pdfs.append(pdf_file)

    if len(new_pdfs) == 0:
        return json_files
    else:
        print("found new pdf's: ", json.dumps(new_pdfs, indent=2))

    scratch_dir = get_scratch_dir()

    old_pdfs = glob.glob(os.path.join(scratch_dir, "*.pdf"))
    for old_pdf in old_pdfs:
        subprocess.call(["rm", f"{old_pdf}"])

    for new_pdf in new_pdfs:
        subprocess.call(["cp", f"{new_pdf}", scratch_dir])

    ds_api, ds_proj = get_ds_api()

    documents = ds.convert_documents(
        api=ds_api, proj_key=ds_proj, source_path=scratch_dir, progress_bar=True
    )

    documents.download_all(result_dir=scratch_dir)
    process_zip_files(scratch_dir)

    info = documents.generate_report(result_dir=scratch_dir)

    for new_pdf in new_pdfs:
        basename = os.path.basename(new_pdf).replace(".pdf", ".json")

        sfile = os.path.join(scratch_dir, basename)
        tfile = new_pdf.replace(".pdf", ".json")

        cmd = ["cp", f"{sfile}", f"{tfile}"]
        print(" ".join(cmd))

        # subprocess.call(cmd)
        subprocess.run(cmd, check=True)
        json_files.append(tfile)

    json_files = sorted(list(set(json_files)))
    return json_files


def ds_list_indices():
    """List all public Deep Search data-indices"""

    api, proj_key = get_ds_api()

    # Fetch list of all data collections
    collections = api.elastic.list()
    collections.sort(key=lambda c: c.name.lower())

    # Visualize summary table
    results = [
        {
            "Name": c.name,
            "Type": c.metadata.type,
            "Num entries": c.documents,
            "Date": c.metadata.created.strftime("%Y-%m-%d"),
            "Id": f"{c.source.elastic_id}",
            "Index": f"{c.source.index_key}",
        }
        for c in collections
    ]
    return results


def create_docs_dir():
    """Create a new documents directory"""

    tdir = get_scratch_dir()

    now = datetime.datetime.now()
    odir = now.strftime("document-collection-%Y-%m-%d_%H-%M-%S")

    odir = os.path.join(tdir, odir)
    return odir


def ds_index_query(
    index,
    query,
    odir=None,
    force=False,
    limit=-1,
    sources=[
        "file-info",
        "description",
        "main-text",
        "texts",
        "tables",
        "figures",
        "page-headers",
        "page-footers",
        "footnotes",
    ],  # Which fields of documents we want to fetch
):
    """Query Deep Search document index"""

    api, proj_key = get_ds_api()

    tdir = get_scratch_dir()

    if odir is None:
        dumpdir = hashlib.md5(f"{index}:{query}".encode()).hexdigest()
        dumpdir = os.path.join(tdir, dumpdir)
    else:
        dumpdir = odir

    if not os.path.exists(dumpdir):
        os.mkdir(dumpdir)
    elif not force:
        return dumpdir

    # Input query
    search_query = f"{query}"  # "\"global warming potential\" AND \"etching\""
    print(f"query: {query}")

    data_collection = ElasticDataCollectionSource(elastic_id="default", index_key=index)
    page_size = 50

    # Prepare the data query
    query = DataQuery(
        search_query,  # The search query to be executed
        source=sources,  # Which fields of documents we want to fetch
        limit=page_size,  # The size of each request page
        coordinates=data_collection,  # The data collection to be queries
    )

    # [Optional] Compute the number of total results matched. This can be used to monitor the pagination progress.
    count_query = copy.deepcopy(query)
    count_query.paginated_task.parameters["limit"] = 0
    count_results = api.queries.run(count_query)

    expected_total = count_results.outputs["data_count"]
    print(f"#-found documents: {expected_total}")

    if limit != -1 and expected_total > limit:
        expected_total = limit

    expected_pages = (
        expected_total + page_size - 1
    ) // page_size  # this is simply a ceiling formula

    # Iterate through all results by fetching `page_size` results at the same time
    bar_format = "{desc:<5.5}{percentage:3.0f}%|{bar:70}{r_bar}"

    count = 0

    all_results = []
    cursor = api.queries.run_paginated_query(query)
    for result_page in tqdm(cursor, total=expected_pages, bar_format=bar_format):
        for row in result_page.outputs["data_outputs"]:
            _id = row["_id"]
            with open(f"{dumpdir}/{_id}.json", "w", encoding="utf-8") as fw:
                fw.write(json.dumps(row["_source"], indent=2))

            count += 1
            if limit != -1 and count >= limit:
                break

        if limit != -1 and count >= limit:
            break

    return dumpdir



#!/usr/bin/env python
"""Module to train name-versus-expr classifier"""

import argparse
import glob
import json
import os
import random
import re

import pandas as pd
from tabulate import tabulate
from tqdm import tqdm

from deepsearch_glm.glm_utils import read_edges_in_dataframe, read_nodes_in_dataframe
from deepsearch_glm.nlp_utils import (
    eval_fst,
    init_nlp_model,
    prepare_data_for_fst_training,
    train_fst,
)


def parse_arguments():
    """Function to parse arguments for `nlp_train_semantic`"""

    parser = argparse.ArgumentParser(
        prog="name_classifier", description="script to train name-classifier"
    )

    parser.add_argument(
        "-m",
        "--mode",
        required=True,
        type=str,
        choices=["extract", "prepare", "train", "evaluate", "apply", "test", "all"],
        help="mode",
    )

    parser.add_argument(
        "-c",
        "--max-count",
        required=True,
        type=int,
        default=-1,
        help="mode",
    )
    args = parser.parse_args()

    return args.mode, int(args.max_count)


def extract(idir: str, ifile: str, max_count: int = -1):
    """Function to extract training data from meta-data of documents"""

    model = init_nlp_model("name")

    fnames = sorted(glob.glob(os.path.join(idir, "*.json")))

    items = []

    for fname in tqdm(fnames, total=len(fnames), ncols=140):
        with open(fname, "r") as fr:
            doc = json.load(fr)

        desc = doc["description"]

        title = desc.get("title", "").strip()

        subjects = desc.get("subjects", [])

        authors = desc.get("authors", [])
        affls = desc.get("affiliations", [])

        # topics = {}

        if False and len(title) > 0:
            res = model.apply_on_text(title)

            insts = pd.DataFrame(
                res["instances"]["data"], columns=res["instances"]["headers"]
            )

            for i, inst in insts.iterrows():
                if inst["type"] == "term":
                    val = random.random()
                    items.append(
                        {
                            "label": "term",
                            "training-sample": (val < 0.9),
                            "text": inst["original"],
                        }
                    )

        for _ in authors:
            name = _["name"]

            label = "person-name"

            val = random.random()
            items.append({"label": label, "training-sample": (val < 0.9), "text": name})

        for _ in affls:
            text = _["name"]

            res = model.apply_on_text(text)

            val = random.random()
            items.append(
                {
                    "label": "term",  # inst["subtype"],
                    "training-sample": (val < 0.9),
                    "text": text,
                }
            )

            if "instances" in res:
                insts = pd.DataFrame(
                    res["instances"]["data"], columns=res["instances"]["headers"]
                )

                for i, inst in insts.iterrows():
                    if inst["type"] == "name":  # and inst["subtype"]!="person-name":
                        val = random.random()
                        items.append(
                            {
                                "label": "term",  # inst["subtype"],
                                "training-sample": (val < 0.9),
                                "text": inst["original"],
                            }
                        )

        if max_count > 0 and len(items) > max_count:
            break

    # print(json.dumps(topics, indent=2))

    print("#-items: ", len(items))

    random.shuffle(items)

    fout = open(ifile, "w")
    for item in tqdm(items, total=len(items), ncols=140):
        fout.write(json.dumps(item) + "\n")
    fout.close()


def prepare(fname):
    """Prepare and normalise data to train classifier"""

    prepare_data_for_fst_training(data_file=fname, loglevel="INFO")


def train(ifile: str):
    """Train FST classifier"""

    train_fst(
        data_file=ifile,
        model_file=f"{ifile}.v2.bin",
        metrics_file=f"{ifile}.v2.metrics.txt",
        ngram=0,
        duration=3600,
        modelsize="10M",
        loglevel="INFO",
    )


def evaluate(ifile: str):
    """Evaluate FST classifier"""

    eval_fst(
        data_file=ifile,
        model_file=f"{ifile}.v2.bin",
        metrics_file=f"{ifile}.v2.metrics.txt",
        loglevel="INFO",
    )


def apply_model(ifile):
    """Apply FST model"""

    model = init_nlp_model(f"custom_fst(name:{ifile}.v2.bin)")

    nodes = pd.read_csv("./data_names/data_abstracts_expr.csv")
    # nodes = pd.read_csv("./data_names/data_abstracts_person-name.csv")
    # nodes = pd.read_csv("./data_names/data_authors_person-name.csv")

    for i, row in nodes.iterrows():
        text = row["text"]

        try:
            res = model.apply_on_text(text)
            # print(res)

            props = pd.DataFrame(
                res["properties"]["data"], columns=res["properties"]["headers"]
            )

            label = props.loc[0, "label"]
            conf = int(100 * props.loc[0, "confidence"])

            # if label!="expr" or conf<=90:
            # if "." in text:
            if label != "term":
                print(f"{i}\t{conf}\t{label}\t{text}")
            # print(props)
            # if i>10000:
            #    break
        except:
            print(f"SKIPPING: {text}")
            continue


def test(ifile):
    """Test FST model"""

    model = init_nlp_model(f"custom_fst(name:{ifile}.v2.bin)")

    # nodes = pd.read_csv("./data_names/data_abstracts_person-name.csv")

    while True:
        text = input("text: ")

        res = model.apply_on_text(text)

        props = pd.DataFrame(
            res["properties"]["data"], columns=res["properties"]["headers"]
        )

        print("\n", props, "\n")


if __name__ == "__main__":
    idir = "./data_personnames"
    ifile = f"./{idir}/training_data.jsonl"

    mode, max_count = parse_arguments()

    if mode == "extract" or mode == "all":
        extract(idir=f"./{idir}/pubmed_meta/", ifile=ifile, max_count=max_count)

    if mode == "prepare" or mode == "all":
        prepare(ifile)

    if mode == "train" or mode == "all":
        train(ifile)

    if mode == "evaluate" or mode == "all":
        evaluate(ifile)

    if mode == "apply":
        apply_model(ifile)

    if mode == "test":
        test(ifile)



import re
from pathlib import Path
from typing import List

import pandas as pd
from docling_core.types.doc import (
    BoundingBox,
    CoordOrigin,
    DocItemLabel,
    DoclingDocument,
    DocumentOrigin,
    GroupLabel,
    ProvenanceItem,
    Size,
    TableCell,
    TableData,
)


def resolve_item(paths, obj):
    """Find item in document from a reference path"""

    if len(paths) == 0:
        return obj

    if paths[0] == "#":
        return resolve_item(paths[1:], obj)

    try:
        key = int(paths[0])
    except:
        key = paths[0]

    if len(paths) == 1:
        if isinstance(key, str) and key in obj:
            return obj[key]
        elif isinstance(key, int) and key < len(obj):
            return obj[key]
        else:
            return None

    elif len(paths) > 1:
        if isinstance(key, str) and key in obj:
            return resolve_item(paths[1:], obj[key])
        elif isinstance(key, int) and key < len(obj):
            return resolve_item(paths[1:], obj[key])
        else:
            return None

    else:
        return None


def _flatten_table_grid(grid: List[List[dict]]) -> List[dict]:
    unique_objects = []
    seen_spans = set()

    for sublist in grid:
        for obj in sublist:
            # Convert the spans list to a tuple of tuples for hashing
            spans_tuple = tuple(tuple(span) for span in obj["spans"])
            if spans_tuple not in seen_spans:
                seen_spans.add(spans_tuple)
                unique_objects.append(obj)

    return unique_objects


def to_docling_document(doc_glm, update_name_label=False) -> DoclingDocument:
    origin = DocumentOrigin(
        mimetype="application/pdf",
        filename=doc_glm["file-info"]["filename"],
        binary_hash=doc_glm["file-info"]["document-hash"],
    )
    doc_name = Path(origin.filename).stem

    doc: DoclingDocument = DoclingDocument(name=doc_name, origin=origin)

    if "properties" in doc_glm:
        props = pd.DataFrame(
            doc_glm["properties"]["data"], columns=doc_glm["properties"]["headers"]
        )
    else:
        props = pd.DataFrame()

    current_list = None

    for ix, pelem in enumerate(doc_glm["page-elements"]):
        ptype = pelem["type"]
        span_i = pelem["span"][0]
        span_j = pelem["span"][1]

        if "iref" not in pelem:
            # print(json.dumps(pelem, indent=2))
            continue

        iref = pelem["iref"]

        if re.match("#/figures/(\\d+)/captions/(.+)", iref):
            # print(f"skip {iref}")
            continue

        if re.match("#/tables/(\\d+)/captions/(.+)", iref):
            # print(f"skip {iref}")
            continue

        path = iref.split("/")
        obj = resolve_item(path, doc_glm)

        if obj is None:
            current_list = None
            print(f"warning: undefined {path}")
            continue

        if ptype == "figure":
            current_list = None
            text = ""
            caption_refs = []
            for caption in obj["captions"]:
                text += caption["text"]

                for nprov in caption["prov"]:
                    npaths = nprov["$ref"].split("/")
                    nelem = resolve_item(npaths, doc_glm)

                    if nelem is None:
                        # print(f"warning: undefined caption {npaths}")
                        continue

                    span_i = nelem["span"][0]
                    span_j = nelem["span"][1]

                    cap_text = caption["text"][span_i:span_j]

                    # doc_glm["page-elements"].remove(nelem)

                    prov = ProvenanceItem(
                        page_no=nelem["page"],
                        charspan=tuple(nelem["span"]),
                        bbox=BoundingBox.from_tuple(
                            nelem["bbox"], origin=CoordOrigin.BOTTOMLEFT
                        ),
                    )

                    caption_obj = doc.add_text(
                        label=DocItemLabel.CAPTION, text=cap_text, prov=prov
                    )
                    caption_refs.append(caption_obj.get_ref())

            prov = ProvenanceItem(
                page_no=pelem["page"],
                charspan=(0, len(text)),
                bbox=BoundingBox.from_tuple(
                    pelem["bbox"], origin=CoordOrigin.BOTTOMLEFT
                ),
            )

            pic = doc.add_picture(prov=prov)
            pic.captions.extend(caption_refs)

        elif ptype == "table":
            current_list = None
            text = ""
            caption_refs = []
            for caption in obj["captions"]:
                text += caption["text"]

                for nprov in caption["prov"]:
                    npaths = nprov["$ref"].split("/")
                    nelem = resolve_item(npaths, doc_glm)

                    if nelem is None:
                        # print(f"warning: undefined caption {npaths}")
                        continue

                    span_i = nelem["span"][0]
                    span_j = nelem["span"][1]

                    cap_text = caption["text"][span_i:span_j]

                    # doc_glm["page-elements"].remove(nelem)

                    prov = ProvenanceItem(
                        page_no=nelem["page"],
                        charspan=tuple(nelem["span"]),
                        bbox=BoundingBox.from_tuple(
                            nelem["bbox"], origin=CoordOrigin.BOTTOMLEFT
                        ),
                    )

                    caption_obj = doc.add_text(
                        label=DocItemLabel.CAPTION, text=cap_text, prov=prov
                    )
                    caption_refs.append(caption_obj.get_ref())

            table_cells_glm = _flatten_table_grid(obj["data"])

            table_cells = []
            for tbl_cell_glm in table_cells_glm:
                if tbl_cell_glm["bbox"] is not None:
                    bbox = BoundingBox.from_tuple(
                        tbl_cell_glm["bbox"], origin=CoordOrigin.BOTTOMLEFT
                    )
                else:
                    bbox = None

                is_col_header = False
                is_row_header = False
                is_row_section = False

                if tbl_cell_glm["type"] == "col_header":
                    is_col_header = True
                elif tbl_cell_glm["type"] == "row_header":
                    is_row_header = True
                elif tbl_cell_glm["type"] == "row_section":
                    is_row_section = True

                table_cells.append(
                    TableCell(
                        row_span=tbl_cell_glm["row-span"][1]
                        - tbl_cell_glm["row-span"][0],
                        col_span=tbl_cell_glm["col-span"][1]
                        - tbl_cell_glm["col-span"][0],
                        start_row_offset_idx=tbl_cell_glm["row-span"][0],
                        end_row_offset_idx=tbl_cell_glm["row-span"][1],
                        start_col_offset_idx=tbl_cell_glm["col-span"][0],
                        end_col_offset_idx=tbl_cell_glm["col-span"][1],
                        text=tbl_cell_glm["text"],
                        bbox=bbox,
                        column_header=is_col_header,
                        row_header=is_row_header,
                        row_section=is_row_section,
                    )
                )

            tbl_data = TableData(
                num_rows=obj.get("#-rows", 0),
                num_cols=obj.get("#-cols", 0),
                table_cells=table_cells,
            )

            prov = ProvenanceItem(
                page_no=pelem["page"],
                charspan=(0, 0),
                bbox=BoundingBox.from_tuple(
                    pelem["bbox"], origin=CoordOrigin.BOTTOMLEFT
                ),
            )

            tbl = doc.add_table(data=tbl_data, prov=prov)
            tbl.captions.extend(caption_refs)

        elif "text" in obj:
            text = obj["text"][span_i:span_j]

            type_label = pelem["type"]
            name_label = pelem["name"]
            if update_name_label and len(props) > 0 and type_label == "paragraph":
                prop = props[
                    (props["type"] == "semantic") & (props["subj_path"] == iref)
                ]
                if len(prop) == 1 and prop.iloc[0]["confidence"] > 0.85:
                    name_label = prop.iloc[0]["label"]

            prov = ProvenanceItem(
                page_no=pelem["page"],
                charspan=(0, len(text)),
                bbox=BoundingBox.from_tuple(
                    pelem["bbox"], origin=CoordOrigin.BOTTOMLEFT
                ),
            )
            label = DocItemLabel(name_label)

            if label == DocItemLabel.LIST_ITEM:
                if current_list is None:
                    current_list = doc.add_group(label=GroupLabel.LIST, name="list")

                # TODO: Infer if this is a numbered or a bullet list item
                doc.add_list_item(
                    text=text, enumerated=False, prov=prov, parent=current_list
                )
            elif label == DocItemLabel.SECTION_HEADER:
                current_list = None

                doc.add_heading(text=text, prov=prov)
            else:
                current_list = None

                doc.add_text(label=DocItemLabel(name_label), text=text, prov=prov)

    for page_dim in doc_glm["page-dimensions"]:
        page_no = int(page_dim["page"])
        size = Size(width=page_dim["width"], height=page_dim["height"])

        doc.add_page(page_no=page_no, size=size)

    return doc


def to_legacy_document_format(doc_glm, doc_leg={}, update_name_label=False):
    """Convert Document object (with `body`) to its legacy format (with `main-text`)"""

    reverse_label_mapping = {
        DocItemLabel.CAPTION.value: "caption",
        DocItemLabel.FOOTNOTE.value: "footnote",
        DocItemLabel.FORMULA.value: "formula",
        DocItemLabel.LIST_ITEM.value: "list-item",
        DocItemLabel.PAGE_FOOTER.value: "page-footer",
        DocItemLabel.PAGE_HEADER.value: "page-header",
        DocItemLabel.PICTURE.value: "picture",  # low threshold adjust to capture chemical structures for examples.
        DocItemLabel.SECTION_HEADER.value: "section-header",
        DocItemLabel.TABLE.value: "table",
        DocItemLabel.TEXT.value: "text",
        DocItemLabel.TITLE.value: "title",
        DocItemLabel.DOCUMENT_INDEX.value: "document index",
        DocItemLabel.CODE.value: "code",
        DocItemLabel.CHECKBOX_SELECTED.value: "checkbox-selected",
        DocItemLabel.CHECKBOX_UNSELECTED.value: "checkbox-unselected",
        DocItemLabel.FORM.value: "form",
        DocItemLabel.KEY_VALUE_REGION.value: "key-value region",
        DocItemLabel.PARAGRAPH.value: "paragraph",
        "subtitle-level-1": "subtitle-level-1",
    }

    extra_mappings = {}
    for v in reverse_label_mapping.values():
        extra_mappings[v] = v
        # extra_mappings[v.lower()] = v
    reverse_label_mapping = {**reverse_label_mapping, **extra_mappings}

    layout_label_to_ds_type = {
        DocItemLabel.TITLE: "title",
        DocItemLabel.DOCUMENT_INDEX: "table-of-contents",
        DocItemLabel.SECTION_HEADER: "subtitle-level-1",
        DocItemLabel.CHECKBOX_SELECTED: "checkbox-selected",
        DocItemLabel.CHECKBOX_UNSELECTED: "checkbox-unselected",
        DocItemLabel.CAPTION: "caption",
        DocItemLabel.PAGE_HEADER: "page-header",
        DocItemLabel.PAGE_FOOTER: "page-footer",
        DocItemLabel.FOOTNOTE: "footnote",
        DocItemLabel.TABLE: "table",
        DocItemLabel.FORMULA: "equation",
        DocItemLabel.LIST_ITEM: "paragraph",
        DocItemLabel.CODE: "paragraph",
        DocItemLabel.PICTURE: "figure",
        DocItemLabel.TEXT: "paragraph",
        DocItemLabel.PARAGRAPH: "paragraph",
    }
    extra_mappings = {}
    for v in layout_label_to_ds_type.values():
        # extra_mappings[v[:1].upper() + v[1:]] = v # capitalize
        extra_mappings[v] = v
    layout_label_to_ds_type = {**layout_label_to_ds_type, **extra_mappings}

    doc_leg["main-text"] = []
    doc_leg["figures"] = []
    doc_leg["tables"] = []
    doc_leg["page-headers"] = []
    doc_leg["page-footers"] = []
    doc_leg["footnotes"] = []
    doc_leg["equations"] = []

    if "properties" in doc_glm:
        props = pd.DataFrame(
            doc_glm["properties"]["data"], columns=doc_glm["properties"]["headers"]
        )
    else:
        props = pd.DataFrame()

    for pelem in doc_glm["page-elements"]:
        ptype = pelem["type"]
        span_i = pelem["span"][0]
        span_j = pelem["span"][1]

        if "iref" not in pelem:
            # print(json.dumps(pelem, indent=2))
            continue

        iref = pelem["iref"]

        if re.match("#/figures/(\\d+)/captions/(.+)", iref):
            # print(f"skip {iref}")
            continue

        if re.match("#/tables/(\\d+)/captions/(.+)", iref):
            # print(f"skip {iref}")
            continue

        path = iref.split("/")
        obj = resolve_item(path, doc_glm)

        if obj is None:
            print(f"warning: undefined {path}")
            continue

        if ptype == "figure":
            text = ""
            for caption in obj["captions"]:
                text += caption["text"]

                for nprov in caption["prov"]:
                    npaths = nprov["$ref"].split("/")
                    nelem = resolve_item(npaths, doc_glm)

                    if nelem is None:
                        print(f"warning: undefined caption {npaths}")
                        continue

                    span_i = nelem["span"][0]
                    span_j = nelem["span"][1]

                    text = caption["text"][span_i:span_j]

                    pitem = {
                        "text": text,
                        "name": reverse_label_mapping[nelem["name"]],
                        "type": layout_label_to_ds_type[nelem["type"]],
                        "prov": [
                            {
                                "bbox": nelem["bbox"],
                                "page": nelem["page"],
                                "span": [0, len(text)],
                            }
                        ],
                    }
                    doc_leg["main-text"].append(pitem)

            find = len(doc_leg["figures"])

            figure = {
                "confidence": obj.get("confidence", 0),
                "created_by": obj.get("created_by", ""),
                "type": obj.get("type", "figure"),
                "cells": [],
                "data": [],
                "text": text,
                "prov": [
                    {
                        "bbox": pelem["bbox"],
                        "page": pelem["page"],
                        "span": [0, len(text)],
                    }
                ],
            }
            doc_leg["figures"].append(figure)

            pitem = {
                "$ref": f"#/figures/{find}",
                "name": reverse_label_mapping[pelem["name"]],
                "type": pelem["type"],
            }
            doc_leg["main-text"].append(pitem)

        elif ptype == "table":
            text = ""
            for caption in obj["captions"]:
                text += caption["text"]

                for nprov in caption["prov"]:
                    npaths = nprov["$ref"].split("/")
                    nelem = resolve_item(npaths, doc_glm)

                    if nelem is None:
                        print(f"warning: undefined caption {npaths}")
                        continue

                    span_i = nelem["span"][0]
                    span_j = nelem["span"][1]

                    text = caption["text"][span_i:span_j]

                    pitem = {
                        "text": text,
                        "name": reverse_label_mapping[nelem["name"]],
                        "type": layout_label_to_ds_type[nelem["type"]],
                        "prov": [
                            {
                                "bbox": nelem["bbox"],
                                "page": nelem["page"],
                                "span": [0, len(text)],
                            }
                        ],
                    }
                    doc_leg["main-text"].append(pitem)

            tind = len(doc_leg["tables"])

            table = {
                "#-cols": obj.get("#-cols", 0),
                "#-rows": obj.get("#-rows", 0),
                "confidence": obj.get("confidence", 0),
                "created_by": obj.get("created_by", ""),
                "type": obj.get("type", "table"),
                "cells": [],
                "data": obj["data"],
                "text": text,
                "prov": [
                    {"bbox": pelem["bbox"], "page": pelem["page"], "span": [0, 0]}
                ],
            }
            doc_leg["tables"].append(table)

            pitem = {
                "$ref": f"#/tables/{tind}",
                "name": reverse_label_mapping[pelem["name"]],
                "type": pelem["type"],
            }
            doc_leg["main-text"].append(pitem)

        elif "text" in obj:
            text = obj["text"][span_i:span_j]

            type_label = layout_label_to_ds_type[pelem["type"]]
            name_label = reverse_label_mapping[pelem["name"]]
            if update_name_label and len(props) > 0 and type_label == "paragraph":
                prop = props[
                    (props["type"] == "semantic") & (props["subj_path"] == iref)
                ]
                if len(prop) == 1 and prop.iloc[0]["confidence"] > 0.85:
                    name_label = prop.iloc[0]["label"]

            pitem = {
                "text": text,
                "name": name_label,  # pelem["name"],
                "type": type_label,  # pelem["type"],
                "prov": [
                    {
                        "bbox": pelem["bbox"],
                        "page": pelem["page"],
                        "span": [0, len(text)],
                    }
                ],
            }
            doc_leg["main-text"].append(pitem)

        else:
            pitem = {
                "name": reverse_label_mapping[pelem["name"]],
                "type": layout_label_to_ds_type[pelem["type"]],
                "prov": [
                    {"bbox": pelem["bbox"], "page": pelem["page"], "span": [0, 0]}
                ],
            }
            doc_leg["main-text"].append(pitem)

    return doc_leg


def to_xml_format(doc_glm, normalised_pagedim: int = -1):
    result = "<document>\n"

    page_dims = pd.DataFrame()
    if "page-dimensions":
        page_dims = pd.DataFrame(doc_glm["page-dimensions"])

    for pelem in doc_glm["page-elements"]:
        ptype = pelem["type"]
        span_i = pelem["span"][0]
        span_j = pelem["span"][1]

        if "iref" not in pelem:
            # print(json.dumps(pelem, indent=2))
            continue

        iref = pelem["iref"]

        page = pelem["page"]
        bbox = pelem["bbox"]

        x0 = bbox[0]
        y0 = bbox[1]
        x1 = bbox[2]
        y1 = bbox[3]

        if normalised_pagedim > 0 and len(page_dims[page_dims["page"] == page]) > 0:
            page_width = page_dims[page_dims["page"] == page].iloc[0]["width"]
            page_height = page_dims[page_dims["page"] == page].iloc[0]["height"]

            rx0 = float(x0) / float(page_width) * normalised_pagedim
            rx1 = float(x1) / float(page_width) * normalised_pagedim

            ry0 = float(y0) / float(page_height) * normalised_pagedim
            ry1 = float(y1) / float(page_height) * normalised_pagedim

            x0 = max(0, min(normalised_pagedim, round(rx0)))
            x1 = max(0, min(normalised_pagedim, round(rx1)))

            y0 = max(0, min(normalised_pagedim, round(ry0)))
            y1 = max(0, min(normalised_pagedim, round(ry1)))

        elif normalised_pagedim > 0 and len(page_dims[page_dims["page"] == page]) == 0:
            print(f"ERROR: no page dimensions for page {page}")

        if re.match("#/figures/(\\d+)/captions/(.+)", iref):
            # print(f"skip {iref}")
            continue

        if re.match("#/tables/(\\d+)/captions/(.+)", iref):
            # print(f"skip {iref}")
            continue

        path = iref.split("/")
        obj = resolve_item(path, doc_glm)

        if obj is None:
            print(f"warning: undefined {path}")
            continue

        if ptype == "figure":
            result += f"<figure bbox=[{x0}, {y0}, {x1}, {y1}]></figure>\n"

        elif ptype == "table":
            result += f"<table bbox=[{x0}, {y0}, {x1}, {y1}]></table>\n"

        elif "text" in obj:
            text = obj["text"][span_i:span_j]
            text_type = pelem["type"]

            result += (
                f"<{text_type} bbox=[{x0}, {y0}, {x1}, {y1}]>{text}</{text_type}>\n"
            )
        else:
            continue

    result += "</document>"

    return result



#!/usr/bin/env python
"""Module providing utility functions for GLM"""

# import argparse
import datetime

# import glob
# import json
import os
import textwrap
from typing import List

import pandas as pd
from tabulate import tabulate

from deepsearch_glm.andromeda_glm import glm_model, glm_query
from deepsearch_glm.utils.common import get_scratch_dir


def create_glm_dir():
    """Function creating directory to write GLM-files."""

    tdir = get_scratch_dir()

    now = datetime.datetime.now()
    glmdir = now.strftime("GLM-model-%Y-%m-%d_%H-%M-%S")

    odir = os.path.join(tdir, glmdir)
    return odir


def load_glm_config(idir: str):
    """Function to create config for loading GLM."""

    config = {
        "IO": {
            "load": {"root": idir},
        },
    }

    return config


def load_glm(idir: str, loglevel: str = "WARNING"):
    """Function to load GLM"""

    config = load_glm_config(idir)

    glm = glm_model()
    glm.set_loglevel(loglevel)

    glm.load(config)

    return glm


def read_nodes_in_dataframe(node_file: str):
    """Function to read nodes from a GLM into a dataframe"""

    df = None

    if node_file.endswith(".csv") and os.path.exists(node_file):
        df = pd.read_csv(
            node_file,
            dtype={
                "hash": "UInt64",
                "total-count": "UInt64",
                "sentence-count": "UInt64",
                "text-count": "UInt64",
                "table-count": "UInt64",
                "document-count": "UInt64",
            },
        )

    return df


def read_edges_in_dataframe(edge_file: str):
    """Function to read edges from a GLM into a dataframe"""

    df = None

    if edge_file.endswith(".csv") and os.path.exists(edge_file):
        df = pd.read_csv(
            edge_file, dtype={"hash": "UInt64", "hash_i": "UInt64", "hash_j": "UInt64"}
        )

    return df


def create_glm_from_config(config: dict, loglevel: str = "WARNING"):
    """Function to create config to create GLM"""

    glm = glm_model()
    glm.set_loglevel(loglevel)

    glm.create(config)

    return config["IO"]["save"]["root"], glm


def create_glm_config_from_docs(
    odir: str,
    json_files: List[str],
    nlp_models: str = "conn;verb;term;abbreviation",
    export_csv=True,
):
    """Function to create GLM configuration"""

    config = {
        "IO": {
            "load": {"root": odir},
            "save": {
                "root": odir,
                "write-CSV": export_csv,
                "write-JSON": False,
                "write-path-text": False,
            },
        },
        "create": {
            "enforce-max-size": False,
            "model": {"max-edges": 1e8, "max-nodes": 1e7},
            "number-of-threads": 4,
            "worker": {
                "local-reading-break": True,
                "local-reading-range": [256, 2560],
                "max-edges": 1e7,
                "max-nodes": 1e6,
            },
            "write-nlp-output": False,
        },
        "mode": "create",
        "parameters": {
            "glm-padding": 1,
            "glm-paths": {
                "keep-concatenation": True,
                "keep-connectors": True,
                "keep-terms": True,
                "keep-verbs": True,
                "keep-sentences": True,
                "keep-tables": True,
                "keep-texts": True,
                "keep-docs": True,
            },
            "nlp-models": nlp_models,
        },
        "producers": [
            {
                "input-format": "json",
                "input-paths": json_files,
                "keep-figures": True,
                "keep-tables": True,
                "keep-text": True,
                "order-text": True,
                "output": False,
                "output-format": "nlp.json",
                "subject-type": "DOCUMENT",
            }
        ],
    }

    return config


def create_glm_from_docs(
    odir: str,
    json_files: List[str],
    nlp_models: str = "conn;verb;term;abbreviation",
    export_csv=False,
):
    """Function to create GLM from documents."""

    glm = glm_model()
    glm.set_loglevel("WARNING")

    config = create_glm_config_from_docs(
        odir, json_files, nlp_models, export_csv=export_csv
    )
    glm.create(config)

    return odir, glm


def create_glm_config_from_texts(
    odir: str,
    json_files: List[str],
    nlp_models: str = "conn;verb;term;abbreviation",
    export_csv=True,
):
    """Function to create GLM configuration"""

    config = {
        "IO": {
            "load": {"root": odir},
            "save": {
                "root": odir,
                "write-CSV": export_csv,
                "write-JSON": False,
                "write-path-text": False,
            },
        },
        "create": {
            "enforce-max-size": False,
            "model": {"max-edges": 1e8, "max-nodes": 1e7},
            "number-of-threads": 4,
            "worker": {
                "local-reading-break": True,
                "local-reading-range": [256, 2560],
                "max-edges": 1e7,
                "max-nodes": 1e6,
            },
            "write-nlp-output": False,
        },
        "mode": "create",
        "parameters": {
            "glm-padding": 1,
            "glm-paths": {
                "keep-concatenation": True,
                "keep-connectors": True,
                "keep-terms": True,
                "keep-verbs": True,
                "keep-sentences": True,
                "keep-tables": True,
                "keep-texts": True,
                "keep-docs": True,
            },
            "nlp-models": nlp_models,
        },
        "producers": [
            {
                "input-format": "jsonl",
                "input-max-documents": 1e9,
                "input-paths": json_files,
                "key": "text",
                "output": False,
                "output-format": "annot.jsonl",
                "output-path": os.path.join(odir, "annots"),
                "start-line": 0,
                "subject-type": "TEXT",
            },
        ],
    }

    return config


def create_glm_from_texts(
    odir: str,
    json_files: List[str],
    nlp_models: str = "conn;verb;term;abbreviation",
    export_csv=True,
):
    """Function to create GLM from texts in JSONL-files."""

    glm = glm_model()
    glm.set_loglevel("WARNING")

    config = create_glm_config_from_texts(odir, json_files, nlp_models, export_csv)
    glm.create(config)

    return odir, glm


def show_query_result(res, max_nodes=16):
    """Function to show the result of the query"""

    if ("status" not in res) or (res["status"] != "success"):
        return

    wrapper = textwrap.TextWrapper(width=50)

    print(
        "overview: \n",
        tabulate(res["overview"]["data"], headers=res["overview"]["headers"]),
        "\n",
    )

    for i, item in enumerate(res["result"]):
        headers = item["nodes"]["headers"]
        data = item["nodes"]["data"]

        for j, row in enumerate(data):
            text = row[headers.index("text")]
            data[j][headers.index("text")] = "\n".join(wrapper.wrap(text))

        print(f"operation {i}: \n", tabulate(data[0:max_nodes], headers=headers), "\n")


def expand_terms(glm: glm_model, term: str):
    """Function to expand the terms"""

    # qry = andromeda_glm.glm_query()
    qry = glm_query()

    if " " not in term:
        qry.select({"nodes": [[term]]})
    else:
        qry.select({"nodes": [term.split(" ")]})

    # qry.filter_by({"mode": "node-flavor", "node-flavors": ["word_token"]})
    qry.filter_by({"mode": "node-flavor", "node-flavors": ["term"]})

    # flid = qry.get_last_flid()
    # qry.traverse({"name": "roots", "edge": "to-root"})
    qry.traverse({"name": "tax-up", "edge": "tax-up"})
    # qry.traverse({"edge":"from-root"})
    # qry.traverse({"edge":"from-token"})
    # qry.filter_by({"mode": "node-flavor", "node-flavors":["term"]})

    # qry.filter_by({"mode": "contains", "contains-flid":flid})
    # qry.traverse({"edge":"to-sent"})

    qry.filter_by({"mode": "node-regex", "node-regex": [f".*{term}.*"]})

    config = qry.to_config()
    # print("query: ", json.dumps(config, indent=2))

    res = glm.query(config)
    """
    if "status" in res and res["status"] == "success":
        show_query_result(res)
    else:
        print(res)
        # print(res["status"], ": ", res["message"])
    """

    return res


def propagate(
    nodes_df: pd.DataFrame,
    edges_df: pd.DataFrame,
    node_itype: str,
    node_otype: str,
    node_name: str,
    edge_name: str,
):
    node = nodes_df[(nodes_df["name"] == node_itype) & (nodes_df["text"] == node_name)]

    edges = edges_df[
        (edges_df["name"] == edge_name) & (edges_df["hash_i"].isin(node["hash"]))
    ]

    result = nodes_df[
        (nodes_df["name"] == node_otype) & (nodes_df["hash"].isin(edges["hash_j"]))
    ]

    return result

    """
    print("nodes: ", nodes_file)
    nodes = pd.read_csv(nodes_file)

    print("edges: ", edges_file)
    edges = pd.read_csv(edges_file)

    print(nodes)
    print(edges)

    hist = nodes["name"].value_counts()
    print(hist)
    
    hist = nodes[nodes["name"]=="label"]["text"].value_counts()
    print(hist)

    for subtype in ["expr", "person-name", "person-group", "vau"]:
        print(f"label {subtype}")
        
        label_node = nodes[(nodes["name"]=="label") & (nodes["text"]==subtype)]
        print(label_node)
            
        from_label = edges[(edges["name"]=="from-label") & (edges["hash_i"].isin(label_node["hash"]))]
        print(from_label)
        
        exprs = nodes[(nodes["name"]=="inst") & (nodes["hash"].isin(from_label["hash_j"]))]
        print(exprs)
        
        exprs.to_csv(f"data_names/data_{mode}_{subtype}.csv", index=False)
    """



#!/usr/bin/env python
"""Module to Query/Explore Graph Language Models"""

import argparse

# import datetime
# import glob
# import json
import os
import sys

# from deepsearch_glm.andromeda_glm import glm_model, glm_query
# from deepsearch_glm.andromeda_nlp import nlp_model
# from deepsearch_glm.andromeda_glm import glm_query
# from deepsearch_glm.glm_utils import expand_terms, load_glm, show_query_result
from deepsearch_glm.glm_utils import expand_terms, load_glm

# import textwrap

# from tabulate import tabulate


def parse_arguments():
    """Function to parse arguments for glm_explore"""

    parser = argparse.ArgumentParser(
        prog="glm_explore",
        description="Explore GLM from Deep Search documents",
        epilog="""
examples of execution: 

1 Explore the GLM
    poetry run python ./deepsearch_glm/explore_glm.py --glm-dir <glm-root-dir>

""",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.add_argument(
        "--glm-dir", required=True, type=str, help="root directory of GLM"
    )

    args = parser.parse_args()

    if not os.path.exists(args.glm_dir):
        sys.exit(-1)

    return args.glm_idr


def explore(glm):
    """Function to interactively explore GLM"""

    while True:
        prompt = input("question: ")
        # prompt = "What is the income of IBM in 2022?"
        # prompt = "net-zero"
        if prompt == "q":
            break

        # terms = analyse_prompt(prompt, nlp_model)
        terms = prompt.split(" ")

        expand_terms(terms, glm)


if __name__ == "__main__":
    idir = parse_arguments()

    glm = load_glm(idir)

    explore(glm)



#!/usr/bin/env python
"""Module to load binary files of models and data"""

import json
import os
from pathlib import Path
from typing import Dict, Tuple

import requests


def get_resources_dir():
    """Function to obtain the resources directory"""

    if "DEEPSEARCH_GLM_RESOURCES_DIR" in os.environ:
        resources_dir = os.getenv("DEEPSEARCH_GLM_RESOURCES_DIR")
    else:
        from deepsearch_glm.andromeda_nlp import nlp_model

        model = nlp_model()
        resources_dir = model.get_resources_path()

    return os.path.normpath(resources_dir)


def list_training_data(key: str, force: bool = False, verbose: bool = False):
    """Function to list the training data"""

    return []


def download_items(
    items: Dict[str, Tuple[str, Path]], force: bool = False, verbose: bool = False
) -> Tuple[bool, Dict[str, str]]:
    """
    Iterate through all the items and downloads them.
    The return dictionary will contain the location where items are downloaded.
    If any error occur, the first return value will be false.
    """

    data = {}
    done = True
    for name, (source, target) in items.items():
        if force or (not target.exists()):
            if verbose:
                print(f" -> downloading {name} ... ", end="")

            target.parent.mkdir(exist_ok=True, parents=True)

            with target.open("wb") as fw:
                r = requests.get(source, stream=True)
                if r.ok:
                    for chunk in r.iter_content(chunk_size=8192):
                        fw.write(chunk)
                else:
                    print(f"Error downloading {name}: [{r.status_code}] {r.text}")
                    done = False
                if verbose:
                    print("done!")

            data[name] = target.name
        elif target.exists():
            if verbose:
                print(f" -> already downloaded {name}")
            data[name] = target.name
        else:
            print(f" -> missing {name}")
    return done, data


def load_training_data(
    data_type: str, data_name: str, force: bool = False, verbose: bool = False
):
    """Function to load data to train NLP models"""

    assert data_type in ["text", "crf", "fst"]

    resources_dir = Path(get_resources_dir())

    data_file_path = resources_dir / "data.json"
    with data_file_path.open(encoding="utf-8") as fr:
        training_data = json.load(fr)

    cos_url = training_data["object-store"]
    cos_prfx = training_data["data"]["prefix"]
    cos_path = f"{cos_url}/{cos_prfx}"

    downloads = {}
    for name, files in training_data["data"][data_type].items():
        print(name)
        if name == data_name:
            source = f"{cos_path}/{files[0]}"
            target = resources_dir / str(files[1])

            downloads[name] = (source, target)

    done, data = download_items(downloads)
    return done, data


def load_pretrained_nlp_models(force: bool = False, verbose: bool = False):
    """Function to load pretrained NLP models"""

    resources_dir = Path(get_resources_dir())
    models_file_path = os.path.join(resources_dir, "models.json")

    with open(models_file_path, "r", encoding="utf-8") as fr:
        models = json.load(fr)

    cos_url = models["object-store"]
    cos_prfx = models["nlp"]["prefix"]

    downloads = {}
    for name, files in models["nlp"]["trained-models"].items():
        source = f"{cos_url}/{cos_prfx}/{files[0]}"
        target = resources_dir / files[1]

        downloads[name] = (source, target)

    done, data = download_items(downloads)
    downloaded_models = list(data.keys())
    return downloaded_models



#!/usr/bin/env python

import argparse
import glob
import json
import sys

# import os
from typing import List

import pandas as pd
from tabulate import tabulate

from deepsearch_glm.andromeda_nlp import nlp_model
from deepsearch_glm.utils.doc_utils import to_legacy_document_format, to_xml_format
from deepsearch_glm.utils.ds_utils import convert_pdffiles


def parse_arguments():
    """Parse arguments for `nlp_apply_on_docs`"""

    parser = argparse.ArgumentParser(
        prog="nlp_apply_on_docs",
        description="Apply NLP on `Deep Search` documents",
        epilog="""
examples of execution: 

1.a run on single document (pdf or json) with default model (=`langauge`):

     poetry run python ./deepsearch_glm/nlp_apply_on_docs.py --pdf './data/documents/articles/2305.02334.pdf'
     poetry run python ./deepsearch_glm/nlp_apply_on_docs.py --json './data/documents/articles/2305.02334.json'

1.b run on single document pdf document and enforce conversion (ignore cache):

     poetry run python ./deepsearch_glm/nlp_apply_on_docs.py --pdf './data/documents/articles/2305.02334.pdf' --force-convert True

2. run on multiple documents:

     poetry run python ./deepsearch_glm/nlp_apply_on_docs.py --pdf './data/documents/articles/*.pdf'
     poetry run python ./deepsearch_glm/nlp_apply_on_docs.py --json './data/documents/articles/*.json'

3. run on multiple documents with non-default models:

     poetry run python ./deepsearch_glm/nlp_apply_on_docs.py --pdf './data/documents/articles/2305.*.pdf' --models 'language;term'

""",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.add_argument(
        "--pdf",
        required=False,
        type=str,
        default=None,
        help="filename(s) of pdf document",
    )

    parser.add_argument(
        "--json",
        required=False,
        type=str,
        default=None,
        help="filename(s) of json document",
    )

    parser.add_argument(
        "--models",
        required=False,
        type=str,
        default="language;semantic",
        help="set NLP models (e.g. `language` or `term;sentence`)",
    )

    parser.add_argument(
        "--filters",
        required=False,
        type=str,
        default="",
        help="set output filters (e.g. `properties;instances` only keeps properties and instances), default is no filter",
    )

    parser.add_argument(
        "--force-convert",
        required=False,
        type=bool,
        default=False,
        help="force pdf conversion",
    )

    parser.add_argument(
        "--legacy",
        required=False,
        type=bool,
        default=False,
        help="enforce old legacy format",
    )

    parser.add_argument(
        "--xml",
        required=False,
        type=bool,
        default=False,
        help="enforce xml format",
    )

    args = parser.parse_args()

    pdf = args.pdf
    json = args.json

    if pdf is None and json is None:
        sys.exit(-1)

    if pdf is not None:
        pdf_files = sorted(glob.glob(pdf))
    else:
        pdf_files = []

    if json is not None:
        json_files = sorted(glob.glob(json))
    else:
        json_files = []

    return (
        pdf_files,
        json_files,
        args.models,
        args.filters,
        args.force_convert,
        args.legacy,
        args.xml,
    )


# FIXME: to be replaced with function in nlp_utils
def init_nlp_model(models: str, filters: List[str] = []):
    """Function to initialse NLP models"""

    # model = andromeda_nlp.nlp_model()
    model = nlp_model()

    config = model.get_apply_configs()[0]

    config["models"] = models
    config["subject-filters"] = filters

    model.initialise(config)
    model.set_loglevel("INFO")

    return model


def show_texts(doc_j, props):
    """Function to show the text of the document on shell"""

    data = []
    for item in doc_j["texts"]:
        label = ""
        selection = props[props["subj_hash"] == item["subj_hash"]]
        if len(selection) > 0:
            label = selection.iloc[0]["label"]

        data.append([item["subj_hash"], label, item["text"][0:48]])

    print(tabulate(data, headers=["subj_hash", "label", "text"]))


def show_doc(doc_j):
    """Function to show the document"""

    props = None
    if "properties" in doc_j:
        props = pd.DataFrame(
            doc_j["properties"]["data"], columns=doc_j["properties"]["headers"]
        )
        print("properties: \n\n", props.to_string())

    if "texts" in doc_j:
        show_texts(doc_j, props)

    if "instances" in doc_j:
        inst = pd.DataFrame(
            doc_j["instances"]["data"], columns=doc_j["instances"]["headers"]
        )
        # print("instances: \n\n", inst.to_string())

        meta = inst[inst["type"] == "metadata"]
        print("meta: \n\n", meta)

        """
        terms = inst[inst["type"] == "term"]
        print("terms: \n\n", terms)

        hist = terms["hash"].value_counts()
        for key, val in hist.items():
            name = terms[terms["hash"] == key].iloc[0]["name"]
            print(f"{val}\t{name}")
        """


if __name__ == "__main__":
    (
        pdf_files,
        json_files,
        model_names,
        filters,
        force_convert,
        legacy,
        xml,
    ) = parse_arguments()

    if len(pdf_files) > 0:
        new_json_files = convert_pdffiles(pdf_files, force=force_convert)

        for _ in new_json_files:
            json_files.append(_)

    json_files = sorted(list(set(json_files)))

    filters_list = []
    if len(filters) > 0:
        filters_list = filters.split(";")

    model = init_nlp_model(model_names, filters_list)

    for json_file in json_files:
        print(f"reading {json_file} ... ", end="")
        with open(json_file, "r", encoding="utf-8") as fr:
            doc_i = json.load(fr)

        print("applying models ... ", end="")
        doc_j = model.apply_on_doc(doc_i)

        # print(doc_j.keys())
        show_doc(doc_j)

        nlp_file = json_file.replace(".json", ".nlp.json")
        print(f"writing  models {nlp_file}")

        with open(nlp_file, "w", encoding="utf-8") as fw:
            fw.write(json.dumps(doc_j, indent=2))

        if legacy:
            doc_i = to_legacy_document_format(doc_j, doc_i)

            nlp_file = json_file.replace(".json", ".leg.json")
            print(f"writing  models {nlp_file}")

            with open(nlp_file, "w", encoding="utf-8") as fw:
                fw.write(json.dumps(doc_i, indent=2))

        if xml:
            doc_xmlstr = to_xml_format(doc_j, normalised_pagedim=100)

            nlp_file = json_file.replace(".json", ".xml.json")
            print(f"writing  models {nlp_file}")

            with open(nlp_file, "w", encoding="utf-8") as fw:
                fw.write(doc_xmlstr)



#!/usr/bin/env python
"""Module to create a CRF model (prepare data, train & evaluate)"""

import argparse
import json
import os
import random

import tqdm
from tabulate import tabulate

from deepsearch_glm.nlp_utils import (
    create_nlp_dir,
    eval_crf,
    get_max_items,
    init_nlp_model,
    train_crf,
)


def parse_arguments():
    """Function to parse arguments for `nlp_train_crf`"""

    parser = argparse.ArgumentParser(
        prog="nlp_train_crf",
        description="train CRF model",
        epilog="""
examples of execution: 

1. end-to-end example to train CRF:

    poetry run python ./deepsearch_glm/nlp_train_crf.py -m all --input-file <filename> --output-dir <models-directory>'

2. end-to-end example to train CRF with limited samples:

    poetry run python ./deepsearch_glm/nlp_train_crf.py -m all --input-file <filename> --output-dir <models-directory> --max-items 1000'
""",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.add_argument(
        "-m",
        "--mode",
        required=True,
        default="all",
        help="mode for training CRF model",
        choices=["prepare", "train", "evaluate", "all"],
    )

    parser.add_argument(
        "--input-file",
        required=False,
        type=str,
        default=None,
        help="input-file with annotations in jsonl format",
    )

    parser.add_argument(
        "--output-dir",
        required=False,
        type=str,
        default="./reference-models",
        help="output directory for trained models & prepared data",
    )

    parser.add_argument(
        "--max-items", required=False, type=int, default=-1, help="number of references"
    )

    args = parser.parse_args()

    odir = None
    if args.output_dir is None:
        odir = create_nlp_dir()

    elif not os.path.exists(args.output_dir):
        os.mkdir(args.output_dir)
        odir = args.output_dir

    else:
        odir = args.output_dir

    return args.mode, args.input_file, odir, args.max_items


def annotate_item(
    atem,
    item,
    start_and_end_in_utf8=True,
    debug=False,
):
    """Function to annotate the tokens of the string-item"""

    atem["word_tokens"]["headers"].append("true-label")

    char_i_ind = atem["word_tokens"]["headers"].index("char_i")
    char_j_ind = atem["word_tokens"]["headers"].index("char_j")

    label = "null"
    for ri, row_i in enumerate(atem["word_tokens"]["data"]):
        atem["word_tokens"]["data"][ri].append(label)

    text = atem["text"]

    # print(item["text"])
    for annot in item["annotation"]:
        lbl = annot["label"].replace(" ", "_").lower().strip()
        utf8_rng = [annot["start"], annot["end"]]

        # print(item["text"][utf8_rng[0] : utf8_rng[1]])

        # beg_ustr = item["text"][0:utf8_rng[0]]
        beg_ustr = text[0 : utf8_rng[0]]
        beg_bstr = beg_ustr.encode("utf-8")

        # end_ustr = item["text"][0:utf8_rng[1]]
        end_ustr = text[0 : utf8_rng[1]]
        end_bstr = end_ustr.encode("utf-8")

        byte_rng = [len(beg_bstr), len(end_bstr)]

        if start_and_end_in_utf8:
            char_i = byte_rng[0]
            char_j = byte_rng[1]
        else:
            char_i = utf8_rng[0]
            char_j = utf8_rng[1]

        for ri, row_i in enumerate(atem["word_tokens"]["data"]):
            if char_i <= row_i[char_i_ind] and row_i[char_j_ind] <= char_j:
                atem["word_tokens"]["data"][ri][-1] = lbl

    if debug:
        print(text)
        print(
            "\n\n",
            tabulate(
                atem["word_tokens"]["data"], headers=atem["word_tokens"]["headers"]
            ),
        )

    # btext = text.encode("utf-8")
    # assert len(text)==len(btext)

    atem["annotated"] = True

    return atem


def prepare_crf(rfile: str, ofile: str, max_items: int, ratio: float = 0.9):
    """Function to prepare the data (strings to token-array) with labels"""

    nlp_model = init_nlp_model("language", filters=["properties", "word_tokens"])

    max_items = get_max_items(rfile, max_items)

    fr = open(rfile, "r", encoding="utf-8")
    fw = open(ofile, "w", encoding="utf-8")

    for i in tqdm.tqdm(range(0, max_items)):
        line = fr.readline().strip()
        if line is None or len(line) == 0:
            break

        item = json.loads(line)
        atem = nlp_model.apply_on_text(item["text"])

        atem = annotate_item(atem, item)
        atem["training-sample"] = random.random() < ratio

        if "annotated" in atem and atem["annotated"]:
            fw.write(json.dumps(atem) + "\n")

    fr.close()
    fw.close()


def create_crf_model(mode: str, ifile: str, odir: str, max_items: int):
    """Function to create CRF model"""

    filename = os.path.basename(ifile)

    afile = os.path.join(odir, filename.replace(".jsonl", ".annot.jsonl"))

    crf_model_file = os.path.join(odir, filename.replace(".jsonl", ".crf_model.bin"))
    crf_metrics_file = crf_model_file + ".metrics.txt"

    if mode in ["prepare", "all"]:
        prepare_crf(ifile, afile, max_items, ratio=0.9)

    if mode in ["train", "all"]:
        train_crf(
            # model_name="reference",
            model_name="custom_crf",
            train_file=afile,
            model_file=crf_model_file,
            metrics_file=crf_metrics_file,
        )

    if mode in ["evaluate", "all"]:
        eval_crf(
            # model_name="reference",
            model_name="custom_crf",
            train_file=afile,
            model_file=crf_model_file,
            metrics_file=crf_metrics_file,
        )

    return afile, crf_model_file, crf_metrics_file


if __name__ == "__main__":
    mode, ifile, odir, max_items = parse_arguments()

    create_crf_model(mode, ifile, odir, max_items)



#!/usr/bin/env python

GENERATE = False

import glob
import json
import os

from deepsearch_glm import andromeda_glm
from deepsearch_glm.glm_utils import (
    create_glm_config_from_docs,
    create_glm_config_from_texts,
    create_glm_dir,
    create_glm_from_config,
    create_glm_from_docs,
    create_glm_from_texts,
    expand_terms,
    load_glm,
    read_edges_in_dataframe,
    read_nodes_in_dataframe,
    show_query_result,
)
from deepsearch_glm.utils.load_pretrained_models import (
    load_pretrained_nlp_models,
    load_training_data,
)


def get_dirs(test_name: str):
    # sdir = "./tests/data/glm/test_01A"

    sdir = os.path.join("./tests/data/glm", test_name)
    os.makedirs(sdir, exist_ok=True)

    if GENERATE:
        rdir = os.path.join(sdir, "glm_ref")
        odir = os.path.join(sdir, "glm_ref")
    else:
        rdir = os.path.join(sdir, "glm_ref")
        odir = os.path.join(sdir, "glm_out")

    return sdir, rdir, odir


def test_01A_load_nlp_models():
    """Tests to determine if NLP models are available"""

    models = load_pretrained_nlp_models(force=False, verbose=True)
    print(f"models: {models}")

    assert "language" in models
    assert "semantic" in models
    assert "name" in models
    assert "reference" in models


def test_01B_load_data():
    """Tests to determine if data are available"""

    done, models = load_training_data(
        data_type="text",
        data_name="arxiv-abstracts-2020-Jan",
        force=False,
        verbose=True,
    )

    print(f"models: {models}")
    assert done


def test_02A_create_glm_from_doc():
    """Tests to determine if GLM creation work"""

    sdir, rdir, odir = get_dirs(test_name="test_01A")

    model_names = "spm;semantic;name;conn;verb;term;abbreviation"

    json_files = glob.glob(os.path.join(sdir, "docs/*.json"))

    # config = create_glm_config_from_docs(odir, json_files, model_names)
    # print(json.dumps(config, indent=2))

    glm = create_glm_from_docs(odir, json_files, model_names, export_csv=True)

    with open(os.path.join(rdir, "topology.json")) as fr:
        ref_topo = json.load(fr)

    with open(os.path.join(odir, "topology.json")) as fr:
        out_topo = json.load(fr)

    for i, row_i in enumerate(ref_topo["node-count"]["data"]):
        row_j = out_topo["node-count"]["data"][i]
        # assert row_i == row_j
        if row_i != row_j:
            print(row_i, " != ", row_j)

    for i, row_i in enumerate(ref_topo["edge-count"]["data"]):
        row_j = out_topo["edge-count"]["data"][i]
        # assert row_i == row_j
        if row_i != row_j:
            print(row_i, " != ", row_j)

    assert ref_topo == out_topo


def test_02B_load_glm():
    """Tests to determine if GLM loading work"""

    # idir = "./tests/data/glm/test_01A/glm_out"
    sdir, rdir, odir = get_dirs(test_name="test_01A")

    glm = load_glm(odir)
    out_topo = glm.get_topology()

    with open(os.path.join(odir, "topology.json")) as fr:
        ref_topo = json.load(fr)

    for i, row_i in enumerate(ref_topo["node-count"]["data"]):
        row_j = out_topo["node-count"]["data"][i]
        assert row_i == row_j

    for i, row_i in enumerate(ref_topo["edge-count"]["data"]):
        row_j = out_topo["edge-count"]["data"][i]
        assert row_i == row_j

    assert ref_topo == out_topo


def test_02C_create_glm_from_texts():
    """Tests to determine if GLM creation work from texts"""

    sdir, rdir, odir = get_dirs(test_name="test_02C")

    model_names = "spm;semantic;name;conn;verb;term;abbreviation"

    # json_files = glob.glob(os.path.join(sdir, "docs/*.json"))
    json_files = ["deepsearch_glm/resources/data/text/arxiv-abstracts-2020-Jan.jsonl"]

    config = create_glm_config_from_texts(odir, json_files, model_names)

    # print(json.dumps(config, indent=2))
    config["producers"][0]["key"] = "abstract"

    glm = create_glm_from_config(config)


def test_03A_query_glm():
    """Tests to determine if GLM queries work"""

    # idir = "./tests/data/glm/test_01A/glm_out"
    sdir, rdir, odir = get_dirs(test_name="test_01A")

    nodes = read_nodes_in_dataframe(os.path.join(odir, "nodes.csv"))
    edges = read_edges_in_dataframe(os.path.join(odir, "edges.csv"))

    """
    subw_nodes = nodes[ nodes["name"]=="subw_token"]
    print(subw_nodes)
    
    next_edge = edges[ (edges["hash_i"].isin(subw_nodes["hash"])) & (edges["hash_j"].isin(subw_nodes["hash"])) ]
    print(next_edge)
    #prev_edge = edges[ (edges["name"]=="prev") and (edges["hash_i"] in subw_nodes["hash"]) and (edges["hash_j"] in subw_nodes["hash"]) ]
    """

    glm = load_glm(odir)
    out_topo = glm.get_topology()

    terms = nodes[nodes["name"] == "term"]
    # print("terms: \n", terms)

    cnt = 0
    for i, row in terms.iterrows():
        res = expand_terms(glm, row["nodes-text"])
        show_query_result(res)

        cnt += 1
        if cnt >= 5:
            break



#!/usr/bin/env python
"""Module for NLP utilities"""

# import argparse
import datetime

# import glob
import json
import os
import sys

# import subprocess
import textwrap
from typing import List

import pandas as pd
from rich.console import Console
from tabulate import tabulate

from deepsearch_glm.andromeda_nlp import nlp_model
from deepsearch_glm.utils.common import get_scratch_dir

# import andromeda_nlp


console = Console()


def create_nlp_dir(tdir=None):
    """Function to create NLP directory"""

    if tdir is None:
        tdir = get_scratch_dir()

    now = datetime.datetime.now()
    nlpdir = now.strftime("NLP-model-%Y-%m-%d_%H-%M-%S")

    odir = os.path.join(tdir, nlpdir)
    return odir


def get_max_items(ifile: str, max_lines: int = -1):
    num_lines = sum(1 for _ in open(ifile, "r", encoding="utf-8"))
    if max_lines != -1:
        max_lines = min(max_lines, num_lines)
    else:
        max_lines = num_lines

    return max_lines


def list_nlp_model_configs():
    """Function to list all available NLP models"""

    configs = []

    configs += nlp_model.get_apply_configs()
    configs += nlp_model.get_train_configs()

    return configs


def init_nlp_model(
    model_names: str = "language;term",
    filters: List[str] = [],
    loglevel: str = "WARNING",
):
    """Function to initialise NLP models"""

    # model = andromeda_nlp.nlp_model()
    model = nlp_model()
    model.set_loglevel(loglevel)

    configs = model.get_apply_configs()
    # print(json.dumps(configs, indent=2))

    config = model.get_apply_configs()[0]
    config["models"] = model_names
    config["subject-filters"] = filters

    model.initialise(config)

    return model


def print_key_on_shell(key, items):
    """Function to print NLP-items on shell"""

    df = pd.DataFrame(items["data"], columns=items["headers"])

    if key in ["instances", "entities"]:
        wrapper = textwrap.TextWrapper(width=70)

        df = df[["type", "subtype", "subj_path", "char_i", "char_j", "original"]]

        table = []
        for i, row in df.iterrows():
            _ = []
            for __ in row:
                if isinstance(__, str):
                    _.append("\n".join(wrapper.wrap(__)))
                else:
                    _.append(__)

            table.append(_)

        headers = ["type", "subtype", "subj_path", "char_i", "char_j", "original"]
        console.print(f"{key}: \n", style="yellow")
        console.print(tabulate(table, headers=headers), "\n")

    else:
        df = pd.DataFrame(items["data"], columns=items["headers"])

        console.print(f"{key}: \n", style="yellow")
        console.print(df.to_string(), "\n")


def print_on_shell(text, result):
    """Function to print text on shell"""

    wrapper = textwrap.TextWrapper(width=70)
    console.print(f"\ntext: \n", style="yellow")
    console.print("\n".join(wrapper.wrap(text)), "\n")

    for _ in ["properties", "word-tokens", "instances", "entities", "relations"]:
        if _ in result and len(result[_]["data"]) > 0:
            print_key_on_shell(_, result[_])
        else:
            console.print(f"[yellow]{_}:[/yellow]", " null\n\n")


def extract_metadata_from_doc(doc):
    """Function to extract metadata from document"""

    df = pd.DataFrame(doc["instances"]["data"], columns=doc["instances"]["headers"])

    metadata = df[df["type"] == "metadata"]

    return metadata


def extract_texts_from_doc(doc):
    """Function to extract texts from document"""

    texts = pd.DataFrame.from_records(doc["texts"])

    return texts


def extract_sentences_from_doc(doc):
    """Function to extract the sentences of a document"""

    df = pd.DataFrame(doc["instances"]["data"], columns=doc["instances"]["headers"])

    sents = df[df["type"] == "sentence"]

    return sents


def extract_references_from_doc(doc):
    """Function to extract references from document"""

    texts = pd.DataFrame.from_records(doc["texts"])

    props = pd.DataFrame(
        doc["properties"]["data"], columns=doc["properties"]["headers"]
    )

    insts = pd.DataFrame(doc["instances"]["data"], columns=doc["instances"]["headers"])

    refs = props[props["label"] == "reference"]

    results = []

    for i, ref in refs.iterrows():
        text = texts[texts["hash"] == ref["subj_hash"]]
        refc = insts[insts["subj_hash"] == ref["subj_hash"]]

        results.append(
            {"text": text["text"], "path": text["sref"], "instances": refc.to_records()}
        )

    return results


def train_crf(
    model_name: str,
    train_file: str,
    model_file: str,
    metrics_file: str,
    loglevel: str = "WARNING",
):
    """Function to train CRF model"""

    model = nlp_model()
    model.set_loglevel(loglevel)

    configs = model.get_train_configs()

    was_trained = False
    for config in configs:
        if config["mode"] == "train" and config["model"].startswith("custom_crf"):
            config["files"]["model-name"] = model_name
            config["files"]["model-file"] = model_file
            config["files"]["train-file"] = train_file
            config["files"]["metrics-file"] = metrics_file

            model.train(config)
            was_trained = True

    if not was_trained:
        print(json.dumps(configs, indent=2))
        sys.exit(-1)


def eval_crf(
    model_name: str,
    train_file: str,
    model_file: str,
    metrics_file: str,
    loglevel: str = "WARNING",
):
    """Function to train CRF model"""

    model = nlp_model()
    model.set_loglevel(loglevel)

    configs = model.get_train_configs()

    was_trained = False
    for config in configs:
        if config["mode"] == "train" and config["model"].startswith("custom_crf"):
            config["files"]["model-name"] = model_name
            config["files"]["model-file"] = model_file
            config["files"]["train-file"] = train_file
            config["files"]["metrics-file"] = metrics_file

            model.evaluate(config)
            was_trained = True

    if not was_trained:
        print(json.dumps(configs, indent=2))
        sys.exit(-1)


def train_tok(
    model_type: str,
    model_name: str,
    ifile: str,
    vocab_size: int = 4096,
    control_symbols: List[str] = [],
    user_symbols: List[str] = [],
    loglevel: str = "WARNING",
):
    """Train a new tokenizer model"""

    model = nlp_model()
    model.set_loglevel(loglevel)

    configs = model.get_train_configs()

    was_trained = False
    for config in configs:
        if config["mode"] == "train" and config["model"] == "spm":
            config["args"]["input-file"] = ifile
            config["args"]["model-type"] = model_type
            config["args"]["model-name"] = model_name
            config["args"]["vocab-size"] = vocab_size

            was_trained = model.train(config)
            break

    if not was_trained:
        print(json.dumps(configs, indent=2))
        sys.exit(-1)

    return was_trained, config


# To train a FST model with HPO, one can use
#
# `./fasttext supervised -input <path-to-train.txt> -output model_name -autotune-validation <<path-to-valid.txt>> -autotune-duration 600 -autotune-modelsize 1M`
#
#  => the parameters can be found via
#
# `./fasttext dump model_cooking.bin args`
#
def train_fst_legacy(
    model_name: str, train_file: str, model_file: str, metrics_file: str
):
    """Function to train fasttext model"""

    # model = andromeda_nlp.nlp_model()
    model = nlp_model()

    configs = model.get_train_configs()
    print(configs)

    for config in configs:
        if config["mode"] == "train" and config["model"] == model_name:
            config["files"]["model-file"] = model_file
            config["files"]["train-file"] = train_file
            config["files"]["metrics-file"] = metrics_file

            model.train(config)


def prepare_data_for_fst_training(
    data_file: str,
    loglevel: str = "WARNING"
    # test_file:str, validation_file:str,
    # model_file:str, metrics_file:str,
    # ngram=3, autotune=True, duration=360, modelsize="1M"
):
    """Function to train fasttext classifier"""

    model = nlp_model()

    configs = model.get_train_configs()
    # print(json.dumps(configs, indent=2))

    for config in configs:
        # print(" => ", config["model"])

        if config["mode"] == "train" and config["model"].startswith("custom_fst"):
            config["args"] = {}

            config["files"]["data-file"] = data_file
            model.prepare_data_for_train(config)


# To train a FST model with HPO, one can use
#
# `./fasttext supervised -input <path-to-train.txt> -output model_name -autotune-validation <<path-to-valid.txt>> -autotune-duration 600 -autotune-modelsize 1M`
#
#  => the parameters can be found via
#
# `./fasttext dump model_cooking.bin args`
#
def train_fst(
    # train_file:str, test_file:str, validation_file:str,
    data_file: str,
    model_file: str,
    metrics_file: str,
    autotune=True,
    duration=360,
    modelsize="1M",
    ngram=None,
    loglevel: str = "WARNING",
):
    """Function to train fasttext classifier"""

    model = nlp_model()
    model.set_loglevel(loglevel)

    configs = model.get_train_configs()
    # print(json.dumps(configs, indent=2))

    for config in configs:
        if config["mode"] == "train" and config["model"].startswith("custom_fst"):
            config["args"] = {}
            config["hpo"]["autotune"] = autotune
            config["hpo"]["duration"] = duration
            config["hpo"]["modelsize"] = modelsize

            if ngram != None:
                config["args"]["n-gram"] = ngram

            config["files"]["data-file"] = data_file
            config["files"]["model-file"] = model_file
            config["files"]["metrics-file"] = metrics_file

            # print(json.dumps(config, indent=2))
            model.train(config)


def eval_fst(
    data_file: str, model_file: str, metrics_file: str, loglevel: str = "WARNING"
):
    """Function to evaluate fasttext classifier"""

    model = nlp_model()
    model.set_loglevel(loglevel)

    configs = model.get_train_configs()
    # print(json.dumps(configs, indent=2))

    for config in configs:
        if config["mode"] == "train" and config["model"].startswith("custom_fst"):
            config["args"] = {}

            config["files"]["data-file"] = data_file

            config["files"]["model-file"] = model_file
            config["files"]["metrics-file"] = metrics_file

            model.evaluate(config)



#!/usr/bin/env python

GENERATE = False

import glob
import json
import os

from tabulate import tabulate

from deepsearch_glm.nlp_train_crf import create_crf_model

# from deepsearch_glm.nlp_train_semantic import train_semantic
from deepsearch_glm.nlp_train_tok import create_tok_model
from deepsearch_glm.nlp_utils import (
    eval_crf,
    eval_fst,
    extract_references_from_doc,
    init_nlp_model,
    list_nlp_model_configs,
    prepare_data_for_fst_training,
    train_crf,
    train_fst,
    train_tok,
)
from deepsearch_glm.utils.doc_utils import to_legacy_document_format
from deepsearch_glm.utils.load_pretrained_models import (  # load_pretrained_nlp_data,
    get_resources_dir,
    list_training_data,
    load_pretrained_nlp_models,
    load_training_data,
)


def round_floats(o):
    if isinstance(o, float):
        return round(o, 2)
    if isinstance(o, dict):
        return {k: round_floats(v) for k, v in o.items()}
    if isinstance(o, (list, tuple)):
        return [round_floats(x) for x in o]
    return o


def check_dimensions(item):
    assert "headers" in item
    assert "data" in item

    headers = item["headers"]
    for row in item["data"]:
        assert len(row) == len(headers)


def get_reduced_instances(instances):
    headers = instances["headers"]

    table = []
    for row in instances["data"]:
        if "reference" in row[0] and "texts" in row[4]:
            table.append([row[0], row[1], row[4], row[5], row[-2]])

    return table, [headers[0], headers[1], headers[4], headers[5], headers[-2]]


def compare_docs(doc_i, doc_j):
    for key, val in doc_j.items():
        if isinstance(val, dict) and ("data" in val) and ("headers" in val):
            assert val["headers"] == doc_i[key]["headers"]

            """
            if val["headers"] != doc_i[key]["headers"]:
                return False
            """

            assert len(val["data"]) == len(doc_i[key]["data"])

            """
            if len(val["data"]) != len(doc_i[key]["data"]):
                return False
            """

            for j, row in enumerate(val["data"]):
                """
                if row != doc_i[key]["data"][j]:
                    return False
                """

                if row != doc_i[key]["data"][j]:
                    print(f"1 ({j}): ", row)
                    print(f"2 ({j}): ", doc_i[key]["data"][j])

                assert row == doc_i[key]["data"][j]

        elif doc_i[key] != doc_j[key]:
            assert doc_i[key] == doc_j[key]
            return False

    return True


def test_01_load_nlp_models():
    models = load_pretrained_nlp_models(force=False, verbose=True)
    # print(f"models: {models}")

    assert "language" in models
    assert "semantic" in models
    assert "name" in models
    assert "reference" in models


# _run_nlp_models_on_text():
def test_02A():
    source = "./tests/data/texts/test_02A_text_01.jsonl"
    target = source

    model = init_nlp_model("spm;sentence;language;term")

    sres = model.apply_on_text("FeSe is a material.")
    sres = round_floats(sres)

    if GENERATE:  # generate the test-data
        fw = open(source, "w")
        fw.write(json.dumps(sres) + "\n")
        fw.close()

        assert True

    else:
        with open(target) as fr:
            tres = json.load(fr)
            tres = round_floats(tres)

        for label in ["properties", "instances"]:
            check_dimensions(sres[label])
            assert label in sres

        for label in ["relations"]:
            assert label not in sres

        # print(tres["properties"])
        # print(sres["properties"])

        assert tres == sres


# _run_nlp_models_on_text():
def test_02B():
    source = "./tests/data/texts/test_02B_text_01.jsonl"
    target = source

    filters = ["properties"]

    model = init_nlp_model("sentence;language;term", filters)

    sres = model.apply_on_text("FeSe is a material.")
    sres = round_floats(sres)

    if GENERATE:  # generate the test-data
        fw = open(source, "w")
        fw.write(json.dumps(sres) + "\n")
        fw.close()

        assert True

    else:
        with open(target) as fr:
            tres = json.load(fr)
            tres = round_floats(tres)

        for label in ["text", "properties"]:
            assert label in sres

        for label in ["instances", "relations"]:
            assert label not in sres

        assert tres == sres


# _run_nlp_models_on_document():
def test_03A():
    with open("./tests/data/docs/1806.02284.json") as fr:
        doc = json.load(fr)

    model = init_nlp_model("sentence;language;term;reference;abbreviation")

    res = model.apply_on_doc(doc)
    res = round_floats(res)

    for label in [
        "description",
        "body",
        "meta",
        "page-elements",
        "texts",
        "tables",
        "figures",
        "properties",
        "instances",
        "relations",
    ]:
        assert label in res

    check_dimensions(res["properties"])
    check_dimensions(res["instances"])
    check_dimensions(res["relations"])


# _run_nlp_models_on_document():
def test_03B():
    with open("./tests/data/docs/1806.02284.json") as fr:
        doc = json.load(fr)

    filters = ["applied_models", "properties"]

    model = init_nlp_model("sentence;language;term;reference", filters)

    res = model.apply_on_doc(doc)
    res = round_floats(res)

    for label in [
        "dloc",
        "applied_models",
        "description",
        "body",
        "meta",
        "page-elements",
        "texts",
        "tables",
        "figures",
        "properties",
    ]:
        assert label in res

    for label in ["instances", "relations"]:
        assert label not in res

    check_dimensions(res["properties"])


# _run_nlp_models_on_document():
def test_03C():
    model = init_nlp_model("language;semantic;sentence;term;verb;conn;geoloc;reference")

    source = "./tests/data/docs/1806.02284.json"
    target = "./tests/data/docs/1806.02284.nlp.json"

    if GENERATE:  # generate the test-data
        with open(source) as fr:
            doc = json.load(fr)

        res = model.apply_on_doc(doc)
        res = round_floats(res)

        # extract_references_from_doc(res)

        fw = open(target, "w")
        fw.write(json.dumps(res, indent=2) + "\n")
        fw.close()

        assert True

    else:
        with open(source) as fr:
            sdoc = json.load(fr)

        res = model.apply_on_doc(sdoc)
        res = round_floats(res)

        with open(target) as fr:
            tdoc = json.load(fr)
            tdoc = round_floats(tdoc)

        assert res == tdoc


# run_nlp_models_on_document():
def test_03D():
    model_i = init_nlp_model("term")
    model_j = init_nlp_model("reference")
    # model_j = init_nlp_model("verb")

    model_k = init_nlp_model("term;reference")
    # model_k = init_nlp_model("term;verb")

    source = "./tests/data/docs/1806.02284.json"

    # target_i = "./tests/data/docs/1806.02284.nlp.i.json"
    # target_j = "./tests/data/docs/1806.02284.nlp.j.json"
    # target_k = "./tests/data/docs/1806.02284.nlp.k.json"

    with open(source) as fr:
        doc = json.load(fr)

    # print("apply model_i")
    res_i = model_i.apply_on_doc(doc)
    # res_i = round_floats(res_i)

    """
    fw = open(target_i, "w")
    fw.write(json.dumps(res_i, indent=2) + "\n")
    fw.close()
    """

    # print("apply model_j")
    res_j = model_j.apply_on_doc(res_i)
    # res_j = model_j.apply_on_doc(doc)
    res_j = round_floats(res_j)

    """
    fw = open(target_j, "w")
    fw.write(json.dumps(res_j, indent=2) + "\n")
    fw.close()
    """

    # print("apply model_k")
    res_k = model_k.apply_on_doc(doc)
    res_k = round_floats(res_k)

    """
    fw = open(target_k, "w")
    fw.write(json.dumps(res_k, indent=2) + "\n")
    fw.close()
    """

    assert res_j["tables"] == res_k["tables"]

    """
    print(tabulate(res_j["properties"]["data"][0:30],
    headers=res_j["properties"]["headers"]))
        
    print(tabulate(res_k["properties"]["data"][0:30],
    headers=res_k["properties"]["headers"]))
    """

    assert len(res_j["properties"]["data"]) == len(res_k["properties"]["data"])
    assert res_j["properties"]["data"] == res_k["properties"]["data"]

    table_i, headers_i = get_reduced_instances(res_i["instances"])
    table_j, headers_j = get_reduced_instances(res_j["instances"])
    table_k, headers_k = get_reduced_instances(res_k["instances"])

    # print(tabulate(table_j, headers=headers_j))
    # print(tabulate(table_k, headers=headers_k))

    """
    print("#-inst-i: ", len(table_i))
    print("#-inst-j: ", len(table_j))
    print("#-inst-k: ", len(table_k))
    """
    assert table_j == table_k

    print("#-instances-j: ", len(res_j["instances"]["data"]))
    print("#-instances-k: ", len(res_k["instances"]["data"]))

    # print(tabulate(res_j["instances"]["data"][-30:]))
    # print(tabulate(res_k["instances"]["data"][-30:]))

    for j in range(0, len(res_j["instances"]["data"])):
        found = False
        for k in range(0, len(res_k["instances"]["data"])):
            if res_k["instances"]["data"][k] == res_j["instances"]["data"][j]:
                found = True

        if not found:
            # print(i)
            # print(res_j["instances"]["data"][j])
            print(res_k["instances"]["data"][j])

    for k in range(0, len(res_k["instances"]["data"])):
        found = False
        for j in range(0, len(res_j["instances"]["data"])):
            if res_k["instances"]["data"][k] == res_j["instances"]["data"][j]:
                found = True

        if not found:
            # print(i)
            # print(res_j["instances"]["data"][j])
            print(res_k["instances"]["data"][k])

    assert len(res_j["instances"]["data"]) == len(res_k["instances"]["data"])
    assert res_j["instances"]["data"] == res_k["instances"]["data"]

    assert res_j == res_k


# test term model
def test_04A():
    source = "./tests/data/texts/terms.jsonl"
    target = "./tests/data/texts/terms.nlp.jsonl"

    model = init_nlp_model("language;semantic;sentence;term;verb;conn;geoloc")

    if GENERATE:  # generate the test-data
        with open(source) as fr:
            lines = fr.readlines()

        fw = open(target, "w")

        for line in lines:
            data = json.loads(line)
            data = round_floats(data)

            res = model.apply_on_text(data["text"])
            res = round_floats(res)

            fw.write(json.dumps(res) + "\n")

        fw.close()

    else:
        with open(target) as fr:
            lines = fr.readlines()

        for line in lines:
            data = json.loads(line)
            data = round_floats(data)

            res = model.apply_on_text(data["text"])
            res = round_floats(res)

            """
            for i,row_i in enumerate(res["properties"]["data"]):
                row_j = data["properties"]["data"][i]
                assert row_i==row_j

            for i,row_i in enumerate(res["instances"]["data"]):
                row_j = data["instances"]["data"][i]
                assert row_i==row_j
            """

            assert res == data

    assert True


# test semantic classifier
def test_04B():
    model = init_nlp_model("semantic")

    source = "./tests/data/texts/semantics.jsonl"
    target = "./tests/data/texts/semantics.nlp.jsonl"

    if GENERATE:  # generate the test-data
        with open(source) as fr:
            lines = fr.readlines()

        fw = open(target, "w")

        for line in lines:
            data = json.loads(line)
            data = round_floats(data)

            res = model.apply_on_text(data["text"])
            res = round_floats(res)

            fw.write(json.dumps(res) + "\n")

        fw.close()
        assert True

    else:
        with open(target) as fr:
            lines = fr.readlines()

        for line in lines:
            data = json.loads(line)
            data = round_floats(data)

            res = model.apply_on_text(data["text"])
            res = round_floats(res)

            """
            for i,row_i in enumerate(res["properties"]["data"]):
                row_j = data["properties"]["data"][i]
                assert row_i==row_j
            """

            assert res == data


# test reference model
def test_04C():
    model = init_nlp_model("reference")

    source = "./tests/data/texts/references.jsonl"
    target = "./tests/data/texts/references.nlp.jsonl"

    if GENERATE:  # generate the test-data
        with open(source) as fr:
            lines = fr.readlines()

        fw = open(target, "w")

        for line in lines:
            data = json.loads(line)
            data = round_floats(data)

            res = model.apply_on_text(data["text"])
            res = round_floats(res)

            fw.write(json.dumps(res) + "\n")

        fw.close()
        assert True

    else:
        with open(target) as fr:
            lines = fr.readlines()

        for line in lines:
            data = json.loads(line)
            data = round_floats(data)

            res = model.apply_on_text(data["text"])
            res = round_floats(res)

            assert res == data


def test_05A():
    model = init_nlp_model("reference;term")

    source = "./tests/data/docs/doc_01.old.json"

    target_leg = "./tests/data/docs/doc_01.leg.json"
    target_nlp = "./tests/data/docs/doc_01.nlp.json"

    # print(f"reading {source} ... ", end="")
    with open(source, "r") as fr:
        doc_i = json.load(fr)

    if GENERATE:
        doc_j = model.apply_on_doc(doc_i)
        doc_j = round_floats(doc_j)

        with open(target_nlp, "w") as fw:
            fw.write(json.dumps(doc_j, indent=2))

        doc_i = to_legacy_document_format(doc_j, doc_i)
        doc_i = round_floats(doc_i)

        with open(target_leg, "w") as fw:
            fw.write(json.dumps(doc_i, indent=2))
    else:
        with open(target_nlp, "r") as fr:
            doc_nlp = json.load(fr)
            doc_nlp = round_floats(doc_nlp)

        with open(target_leg, "r") as fr:
            doc_leg = json.load(fr)
            doc_leg = round_floats(doc_leg)

        doc_j = model.apply_on_doc(doc_i)
        doc_j = round_floats(doc_j)

        assert compare_docs(doc_j, doc_nlp)
        """
        for key,val in doc_j.items():

            if isinstance(val, dict) and ("data" in val) and ("headers" in val):
                
                assert val["headers"] == doc_nlp[key]["headers"]
                assert len(val["data"]) == len(doc_nlp[key]["data"])

                for j,row in enumerate(val["data"]):
                    assert row == doc_nlp[key]["data"][j]
                    
            else:
                assert doc_j[key] == doc_nlp[key]
        """

        assert doc_j == doc_nlp

        doc_i = to_legacy_document_format(doc_j, doc_i)
        doc_i = round_floats(doc_i)

        assert compare_docs(doc_i, doc_leg)

        """
        #assert doc_i == doc_leg
        for key,val in doc_i.item():
            assert doc_i[key] == doc_leg[key]
        """


# download CRF data
def test_06A():
    verbose = False

    done, data = load_training_data(
        data_type="crf", data_name="materials", force=False, verbose=verbose
    )
    # done, data = load_pretrained_nlp_data(key="crf", force=False, verbose=verbose)

    if verbose:
        print(json.dumps(data, indent=2))

    assert done


# train CRF
def test_06B():
    resources_dir = get_resources_dir()

    crf_files = glob.glob(f"{resources_dir}/data/nlp/crf.*.jsonl")
    assert len(crf_files) > 0

    # print(crf_files)
    for crf_file in crf_files:
        if crf_file.endswith(".annot.jsonl"):
            continue

        print(f"training on {crf_file}")
        annot_file, model_file, metrics_file = create_crf_model(
            mode="all", ifile=crf_file, odir=os.path.dirname(crf_file), max_items=1000
        )

        assert os.path.exists(annot_file)
        assert os.path.exists(model_file)
        assert os.path.exists(metrics_file)

    assert True


# predict CRF
def test_06C():
    verbose = False

    resources_dir = get_resources_dir()

    # print(f"{resources_dir}/data_nlp/crf.*.jsonl")
    crf_files = glob.glob(f"{resources_dir}/data_nlp/crf.*.jsonl")

    for crf_file in crf_files:
        if crf_file.endswith(".annot.jsonl"):
            continue

        name = os.path.basename(crf_file).replace("crf.", "").replace(".jsonl", "")

        # print(f"running on {crf_file}")
        annot_file, model_file, metrics_file = create_crf_model(
            mode="files", ifile=crf_file, odir=os.path.dirname(crf_file), max_items=1000
        )

        assert os.path.exists(model_file)

        model = init_nlp_model(
            f"language;custom_crf({name}:{model_file})",
            filters=["properties", "instances"],
        )

        fr = open(crf_file, "r")

        while True:
            line = fr.readline()

            if line is None or len(line.strip()) == 0:
                break

            data = json.loads(line)
            sres = model.apply_on_text(data["text"])

            if verbose:
                print(sres["text"])
                print(
                    tabulate(
                        sres["instances"]["data"], headers=sres["instances"]["headers"]
                    )
                )

        fr.close()

    assert True


# download text data for tokenizers
def test_07A():
    verbose = True

    done, data = load_training_data(
        data_type="text",
        data_name="arxiv-abstracts-2020-Jan-txt",
        force=False,
        verbose=verbose,
    )

    if verbose:
        print(json.dumps(data, indent=2))

    assert done


# train tokenizer
def test_07B():
    resources_dir = get_resources_dir()

    txt_file = f"{resources_dir}/data/text/arxiv-abstracts-2020-Jan.txt"
    assert os.path.exists(txt_file)

    model_type = "unigram"
    model_name = "test-tokenizer-model"

    print(f"training on {txt_file}")
    model_file = create_tok_model(
        model_type=model_type, model_name=model_name, ifile=txt_file
    )

    assert os.path.exists(model_name + ".model")
    assert os.path.exists(model_name + ".vocab")


# load custom tokenizer
def test_07C():
    model_name = "test-tokenizer-model"
    model_file = f"{model_name}.model"

    assert os.path.exists(model_file)

    model = init_nlp_model(
        f"language;custom_spm({model_name}:{model_file})",
        filters=["properties", "instances"],
    )


# download text data for fst-classifier
def test_08A():
    verbose = True

    done, data = load_training_data(
        data_type="fst",
        data_name="names-classifier-small",
        force=False,
        verbose=verbose,
    )

    if verbose:
        print(json.dumps(data, indent=2))

    assert done


# train fst-classifier
def test_08B():
    resources_dir = get_resources_dir()

    data_file = f"{resources_dir}/data/nlp/nlp.fst.names-classifier.small.jsonl"
    assert os.path.exists(data_file)

    prepare_data_for_fst_training(data_file=data_file, loglevel="INFO")

    model_file = data_file + ".fst_model.bin"
    metrics_file = data_file + ".fst_metrics.txt"

    train_fst(
        data_file=data_file,
        model_file=model_file,
        metrics_file=metrics_file,
        ngram=3,
        duration=60,
        autotune=True,
        modelsize="1M",
        loglevel="INFO",
    )

    assert os.path.exists(model_file)

    eval_fst(
        data_file=data_file,
        model_file=model_file,
        metrics_file=metrics_file,
        loglevel="INFO",
    )

    assert os.path.exists(metrics_file)



#!/usr/bin/env python
"""Module to train semantic classifier"""

import argparse
import glob
import json
import os
import random
import sys

import pandas as pd
import tqdm
from rich.console import Console
from tabulate import tabulate

from deepsearch_glm.andromeda_nlp import nlp_model
from deepsearch_glm.nlp_utils import create_nlp_dir, init_nlp_model
from deepsearch_glm.utils.ds_utils import ds_index_query

console = Console()


def parse_arguments():
    """Function to parse arguments for `nlp_train_semantic`"""

    parser = argparse.ArgumentParser(
        prog="nlp_train_semantic",
        description="train classifier for semantic text classifier",
        epilog="""
examples of execution: 

1. end-to-end example on pdf documents:

    poetry run python ./deepsearch_glm/nlp_train_semantic.py -m all --input-dir '<root-dir-of-json-docs> --output-dir <models-directory>'

""",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.add_argument(
        "-m",
        "--mode",
        required=True,
        default="all",
        help="mode for training semantic model",
        choices=["retrieve", "prepare", "process", "train", "eval", "refine", "all"],
    )

    parser.add_argument(
        "--input-dir",
        required=True,
        type=str,
        # default="./semantic-models/documents",
        help="input directory with documents",
    )

    parser.add_argument(
        "--output-dir",
        required=False,
        type=str,
        default=None,
        help="output directory for trained models",
    )

    args = parser.parse_args()

    idir = args.input_dir

    if not os.path.exists(idir):
        print(f"input directory {idir} does not exist")
        sys.exit(-1)

    if args.output_dir is None:
        odir = idir

    elif not os.path.exists(args.output_dir):
        os.mkdir(args.output_dir)
        odir = args.output_dir

    else:
        odir = args.output_dir

    return args.mode, idir, odir


def retrieve_data_pubmed(sdir):
    """Function to retrieve data from pubmed folder"""

    tdir = os.path.join(sdir, "pubmed")

    if not os.path.exists(sdir):
        os.mkdir(sdir)

    index = "pubmed"
    query = "description.publication_date:[2022-01-01 TO 2022-03-01]"

    odir = ds_index_query(
        index,
        query,
        tdir,
        sources=["_name", "file-info", "references", "description"],
        force=True,
        limit=1000,
    )

    return odir


def retrieve_data_arxiv(sdir):
    """Function to retrieve data from arxiv folder"""

    tdir = os.path.join(sdir, "arxiv")

    if not os.path.exists(sdir):
        os.mkdir(sdir)

    index = "arxiv"
    query = "description.publication_date:[2022-01-01 TO 2022-03-01]"

    odir = ds_index_query(
        index,
        query,
        tdir,
        sources=["_name", "file-info", "description", "main-text"],
        force=True,
        limit=50000,
    )

    return odir


def retrieve_data(sdir, index):
    """Function to retrieve data"""

    tdir = os.path.join(sdir, index)

    if not os.path.exists(sdir):
        os.mkdir(sdir)

    query = "*"

    odir = ds_index_query(
        index,
        query,
        tdir,
        sources=["_name", "file-info", "description", "main-text"],
        force=True,
        limit=50000,
    )

    return odir


def prepare_data_from_legacy_documents(doc):
    """Function to prepare data from legacy documents"""

    if "file-info" in doc:
        dhash = doc["file-info"]["document-hash"]
    else:
        dhash = -1

    text_len = len(doc["main-text"])

    title_ind = len(doc["main-text"])

    abs_beg = len(doc["main-text"])
    intro_beg = len(doc["main-text"])

    ref_beg = len(doc["main-text"])
    ref_end = len(doc["main-text"])

    data = []
    for i, item in enumerate(doc["main-text"]):
        if "text" not in item:
            continue

        label = item["type"].lower()
        text = item["text"].lower().strip()

        if "title" == label and title_ind == text_len:
            title_ind = i

        if ("title" in label) and ("abstract" in text) and abs_beg == text_len:
            abs_beg = i

        if (text.startswith("abstract")) and abs_beg == text_len:
            abs_beg = i

        if ("title" in label) and ("introduction" in text) and intro_beg == text_len:
            intro_beg = i

        if ("title" in label) and ("references" in text) and ref_beg == text_len:
            ref_beg = i

        # (("title" in label) or ("caption" in label)) and ("reference" not in text):
        if (
            ref_end == text_len
            and ref_beg < text_len
            and i > ref_beg
            and (("title" in label))
            and ("reference" not in text)
        ):
            ref_end = i

    if title_ind == text_len or abs_beg == text_len or ref_beg == text_len:
        return data

    # print(dhash)
    for i, item in enumerate(doc["main-text"]):
        if "text" not in item:
            continue

        type_ = item["type"]
        label = item["type"]
        text = item["text"]

        skip = (
            ((len(text) <= 1) or (len(text.split(" ")) == 1))
            and ("title" not in label)
            and (len(text) <= 5)
        )
        if skip:
            # print(f"skipping: {text}")
            continue

        if title_ind < i and (i < min(abs_beg, intro_beg)):
            label = "meta-data"
        elif ref_beg < i and i < ref_end:
            label = "reference"
        elif "title" in label:
            label = "header"
        else:
            label = "text"

        if "title" in label:
            console.print(f"{label}, {type_}: {text[0:48]}", style="yellow")
        elif "meta" in label:
            console.print(f"\t{label}, {type_}: {text[0:48]}", style="green")
        elif "text" in label:
            console.print(f"\t{label}, {type_}: {text[0:48]}")
        elif "reference" in label:
            console.print(f"\t{label}, {type_}: {text[0:48]}", style="blue")
        else:
            console.print(f"\t{label}, {type_}: {text[0:48]}", style="red")

        if random.random() < 0.9:
            training_sample = True
        else:
            training_sample = False

        data.append(
            {
                "document-hash": dhash,
                "label": label,
                "text": item["text"],
                "training-sample": training_sample,
            }
        )

    return data


def prepare_data_from_description(doc):
    """Function to prepare data from description"""

    if "file-info" in doc:
        dhash = doc["file-info"]["document-hash"]
    else:
        dhash = -1

    data = []

    if "references" in doc:
        for item in doc["references"]:
            data.append(
                {"label": "reference", "text": item["text"], "document-hash": dhash}
            )

    if "description" in doc:
        desc = doc["description"]

        if "title" in desc:
            data.append(
                {"label": "text", "text": desc["title"], "document-hash": dhash}
            )

        if "abstract" in desc:
            for item in desc["abstract"]:
                data.append({"label": "text", "text": item, "document-hash": dhash})

        affiliations = []
        if "affiliations" in desc:
            for item in desc["affiliations"]:
                affiliations.append(item["name"])
                data.append(
                    {"label": "meta-data", "text": item["name"], "document-hash": dhash}
                )

        authors = []
        if "authors" in desc:
            for item in desc["authors"]:
                authors.append(item["name"])
                data.append(
                    {"label": "meta-data", "text": item["name"], "document-hash": dhash}
                )

        if len(authors) > 1:
            data.append(
                {
                    "label": "meta-data",
                    "text": ", ".join(authors),
                    "document-hash": dhash,
                }
            )

        if len(affiliations) > 1:
            data.append(
                {
                    "label": "meta-data",
                    "text": ", ".join(affiliations),
                    "document-hash": dhash,
                }
            )

        if len(authors) >= 1 and len(affiliations) >= 1:
            for _ in authors:
                for __ in affiliations:
                    data.append(
                        {
                            "label": "meta-data",
                            "text": " ".join([_, __]),
                            "document-hash": dhash,
                        }
                    )

    return data


def prepare_data(json_files, data_file):
    """Function to prepare data"""

    num_lines = 0

    fw = open(data_file, "w", encoding="utf-8")

    for json_file in tqdm.tqdm(json_files):
        data = []

        try:
            with open(json_file, "r", encoding="utf-8") as fr:
                doc = json.load(fr)
        except:
            continue

        if "main-text" in doc:
            data = prepare_data_from_legacy_documents(doc)
            # continue
        else:
            data = prepare_data_from_description(doc)

        for item in data:
            num_lines += 1
            fw.write(json.dumps(item) + "\n")
        data = []

    fw.close()

    return num_lines


def process_data(data_file):
    """Function to process data"""

    model = init_nlp_model("semantic")

    configs = model.get_train_configs()
    # print(json.dumps(configs, indent=2))

    for config in configs:
        if config["mode"] == "train" and config["model"] == "semantic":
            config["files"]["train-file"] = data_file
            print(json.dumps(config))

            model.prepare_data_for_train(config)


# To train a FST model with HPO, one can use
#
# `./fasttext supervised -input <path-to-train.txt> -output model_name -autotune-validation <<path-to-valid.txt>> -autotune-duration 600 -autotune-modelsize 1M`
#
#  => the parameters can be found via
#
# `./fasttext dump model_cooking.bin args`
#
def train_fst(
    data_file, model_file, metrics_file, autotune=True, duration=360, modelsize="1M"
):
    """Function to train fasttext classifier"""

    model = nlp_model()

    configs = model.get_train_configs()
    # print(json.dumps(configs, indent=2))

    for config in configs:
        if config["mode"] == "train" and config["model"] == "semantic":
            config["args"] = {}
            config["hpo"]["autotune"] = autotune
            config["hpo"]["duration"] = duration
            config["hpo"]["modelsize"] = modelsize

            config["args"]["n-gram"] = 0

            config["files"]["train-file"] = data_file

            config["files"]["model-file"] = model_file
            config["files"]["metrics-file"] = metrics_file

            model.train(config)


def evaluate_model(data_file, model_file, metrics_file):
    """Function to evaluate fasttext classifier"""

    model = nlp_model()

    configs = model.get_train_configs()
    # print(json.dumps(configs, indent=2))

    for config in configs:
        if config["mode"] == "train" and config["model"] == "semantic":
            config["args"] = {}

            config["files"]["train-file"] = data_file

            config["files"]["model-file"] = model_file + ".bin"
            config["files"]["metrics-file"] = metrics_file

            print(config)
            model.evaluate_model(config)


def refine_data(data_file):
    """Function to refine data"""

    model = init_nlp_model("semantic", filters=["properties"])

    print(f"reading {data_file}")

    fr = open(data_file, "r", encoding="utf-8")
    fw = open(data_file.replace(".jsonl", ".v2.jsonl"), "w", encoding="utf-8")

    table = []

    num_lines = 0
    while True:
        try:
            line = fr.readline()
            data = json.loads(line)
        except:
            break

        num_lines += 1

        res = model.apply_on_text(data["text"])

        pred = res["properties"]["data"][-1][-2]
        conf = res["properties"]["data"][-1][-1]

        if "properties" in res:
            row = [
                data["document-hash"],
                data["training-sample"],
                data["label"],
                conf,
                pred,
                data["text"][0:48],
            ]

            table.append(row)
        else:
            print(json.dumps(res, indent=2))

        if conf > 0.9 or data["label"] == pred:
            fw.write(json.dumps(data) + "\n")

        if num_lines > 100000:
            break

    fr.close()

    print(f"num-lines: {num_lines}")
    print(tabulate(table))

    df = pd.DataFrame(
        table, columns=["document", "training", "true", "conf", "pred", "text"]
    )
    df.sort_values(["pred", "training", "conf"], ascending=[True, True, False])

    df.to_csv(f"{data_file}.csv")


def train_semantic(mode, idir, odir, autotune=True, duration=360, modelsize="1M"):
    """Function to train semantic fasttext classifier"""

    tdir = os.path.join(odir, "documents")

    data_file = os.path.join(odir, "nlp-train-semantic.data.jsonl")
    # annot_file = os.path.join(odir, "nlp-train-semantic.annot.jsonl")

    fst_model_file = os.path.join(odir, "fst_semantic")
    fst_metrics_file = os.path.join(odir, "fst_semantic.metrics.txt")

    if mode in ["all", "retrieve"]:
        retrieve_data_pubmed(tdir)
        json_files = sorted(glob.glob(os.path.join(tdir, "*.json")))

        retrieve_data_arxiv(tdir)
        json_files += sorted(glob.glob(os.path.join(tdir, "*.json")))

    if mode in ["all", "prepare"]:
        json_files = sorted(glob.glob(os.path.join(idir, "*.json")))
        json_files += sorted(glob.glob(os.path.join(idir, "*/*.json")))

        print("#-files: ", len(json_files))
        if len(json_files) == 0:
            exit(-1)

        # json_files = json_files[0:1000]

        num_lines = prepare_data(json_files, data_file)

    if mode in ["all", "process"]:
        process_data(data_file)  # , annot_file)

    if mode in ["all", "train"]:
        # train_fst(annot_file, fst_model_file, fst_metrics_file, autotune=True, duration=300, modelsize="100M")
        train_fst(
            data_file,
            fst_model_file,
            fst_metrics_file,
            autotune=True,
            duration=3600,
            modelsize="100M",
        )

    if mode in ["all", "eval"]:
        evaluate_model(data_file, fst_model_file, fst_metrics_file)

    if mode in ["all", "refine"]:
        refine_data(data_file)


if __name__ == "__main__":
    mode, idir, odir = parse_arguments()

    train_semantic(mode, idir, odir)



#!/usr/bin/env python
"""Module to test the nlp structs from andromeda_nlp"""

import json

import pandas as pd
from tabulate import tabulate

from deepsearch_glm.andromeda_structs import ds_document, ds_table, ds_text
from deepsearch_glm.nlp_utils import init_nlp_model
from deepsearch_glm.utils.load_pretrained_models import load_pretrained_nlp_models

TEXTS = [
    """Anarchism is a political philosophy and movement that is sceptical
of authority and rejects all involuntary, coercive forms of hierarchy. Anarchism
calls for the abolition of the state, which it holds to be unnecessary,
undesirable, and harmful. As a historically left-wing movement, placed on the
farthest left of the political spectrum, it is usually described alongside
communalism and libertarian Marxism as the libertarian wing (libertarian
socialism) of the socialist movement, and has a strong historical association
with anti-capitalism and socialism.""",
    """Humans lived in societies without formal hierarchies long before the
establishment of formal states, realms, or empires. With the rise of organised
hierarchical bodies, scepticism toward authority also rose. Although traces of
anarchist thought are found throughout history, modern anarchism emerged from
the Enlightenment. During the latter half of the 19th and the first decades of
the 20th century, the anarchist movement flourished in most parts of the world
and had a significant role in workers' struggles for emancipation. Various
anarchist schools of thought formed during this period. Anarchists have taken
part in several revolutions, most notably in the Paris Commune, the Russian
Civil War and the Spanish Civil War, whose end marked the end of the classical
era of anarchism. In the last decades of the 20th and into the 21st century,
the anarchist movement has been resurgent once more.""",
]

TABLES = [
    [
        ["Year", "Company", "Revenue"],
        ["2020", "XXX", "10B USD"],
        ["2021", "XXX", "10B USD"],
        ["2022", "XXX", "10B USD"],
        ["2020", "YYY", "12B USD"],
        ["2021", "YYY", "13B USD"],
        ["2022", "YYY", "14B USD"],
    ]
]


def to_dataframe(obj):
    return pd.DataFrame(obj["data"], columns=obj["headers"])


def test_0A_load_nlp_models():
    models = load_pretrained_nlp_models(force=False, verbose=True)

    assert "language" in models
    assert "semantic" in models
    assert "name" in models
    assert "reference" in models


def test_01A():
    subj = ds_text()
    print(subj)

    res = subj.to_json(set([]))
    print(res)

    assert True


def test_01B():
    subj = ds_table()

    res = subj.to_json(set([]))
    print(res)

    assert True


def test_01C():
    subj = ds_document()

    res = subj.to_json(set([]))
    print(res)

    assert True


def test_02A():
    subj = ds_text()

    subj.set_text(TEXTS[0])

    res = subj.to_json(set([]))
    print(res)

    assert True


def test_02B():
    subj = ds_table()

    subj.set_data(TABLES[0])
    res = subj.to_json(set([]))

    assert True


def test_02C():
    subj = ds_document()
    print(subj)

    for text in TEXTS:
        text_subj = ds_text()
        text_subj.set_text(text)

        subj.append_text(text_subj)

    res = subj.to_json(set([]))
    print(res)

    assert True


def test_03A():
    subj = ds_text()
    print(subj)

    subj.set_text(TEXTS[0])

    res = subj.to_json(set([]))
    # print(json.dumps(res, indent=2))

    model = init_nlp_model(
        "language;semantic;sentence;term;verb;conn;geoloc;abbreviation"
    )
    model.apply_on_text(subj)

    res = subj.to_json(set([]))
    print("keys: ", res.keys())

    print(res["applied_models"])

    for key in ["properties", "instances", "relations"]:
        if key in res:
            df = to_dataframe(res[key])
            print(df)

    assert True


def test_03B():
    subj = ds_table()

    subj.set_data(TABLES[0])
    res = subj.to_json(set([]))

    model = init_nlp_model(
        "language;semantic;sentence;term;verb;conn;geoloc;abbreviation"
    )
    model.apply_on_table(subj)

    res = subj.to_json(set([]))
    # res = subj.to_json(set(["word-tokens", "word_tokens", "properties", "instances", "relations"]))
    print("keys: ", res.keys())

    for key in ["properties", "instances", "relations"]:
        if key in res:
            df = to_dataframe(res[key])
            print(df)

    assert True


def test_03C():
    subj = ds_document()

    for text in TEXTS:
        text_subj = ds_text()
        text_subj.set_text(text)

        subj.append_text(text_subj)

    for table in TABLES:
        table_subj = ds_table()
        table_subj.set_data(table)

        subj.append_table(table_subj)

    jdoc = subj.to_json(set())
    print(json.dumps(jdoc, indent=2))

    model = init_nlp_model(
        "language;semantic;sentence;term;verb;conn;geoloc;abbreviation"
    )
    model.apply_on_doc(subj)

    res = subj.to_json(set([]))
    print("keys: ", res.keys())

    for key in ["properties", "instances", "relations"]:
        if key in res:
            df = to_dataframe(res[key])
            print(df)

    assert True



#!/usr/bin/env python
"""Module to do DocQA"""

import argparse

# import glob
import json
import sys

import pandas as pd
from tabulate import tabulate

# from deepsearch_glm.andromeda_glm import glm_model, glm_query
from deepsearch_glm.andromeda_glm import glm_query

# import andromeda_glm
# import andromeda_nlp
# from deepsearch_glm.andromeda_nlp import nlp_model
from deepsearch_glm.glm_utils import load_glm
from deepsearch_glm.nlp_utils import init_nlp_model

# import os
# import textwrap


# from ds_convert import convert_pdffile


def parse_arguments():
    """Function to parse arguments for `glm_docqa`"""

    parser = argparse.ArgumentParser(
        prog="glm_docqa",
        description="Do Q&A on pdf document",
        epilog="Text at the bottom of help",
    )

    parser.add_argument(
        "--glm-dir", required=True, type=str, help="directory of GLM model"
    )

    parser.add_argument(
        "--qa-pairs",
        required=False,
        type=str,
        default="prompt",
        help="CSV file with QA pairs or `prompt`",
    )

    parser.add_argument(
        "--models",
        required=False,
        type=str,
        default="name;verb;term;abbreviation",
        help="set NLP models (e.g. `term;sentence`)",
    )

    parser.add_argument(
        "--show",
        required=False,
        action="store_true",
        help="show a picture of the output",
    )

    args = parser.parse_args()

    return args.glm_dir, args.qa_pairs, args.models, args.show


def analyse_prompt(prompt, nlp_model):
    """Function to analyse the prompt"""

    res = nlp_model.apply_on_text(prompt)

    print(res.keys())
    print(tabulate(res["word-tokens"]["data"], headers=res["word-tokens"]["headers"]))
    print(tabulate(res["entities"]["data"], headers=res["entities"]["headers"]))

    type_ind = res["entities"]["headers"].index("type")
    name_ind = res["entities"]["headers"].index("name")

    terms = []
    for i, row in enumerate(res["entities"]["data"]):
        if row[type_ind] == "term":
            terms.append([row[name_ind]])

    return terms


def compute_topk_on_documents(df, nlp_mdl, glm_mdl, show=False):
    """Function to compute topk of documents"""

    topk = {0: 0}
    for ind in range(1, 10):
        topk[ind] = 0

    for i, row in df.iterrows():
        print(i, row["question"])

        doc_hash = row["doc_hash"]
        question = row["question"]

        context = row["text"]

        qres = nlp_mdl.apply_nlp_on_text(question)
        cres = nlp_mdl.apply_nlp_on_text(context)

        data = cres["instances"]["data"]
        headers = cres["instances"]["headers"]

        insts = []
        for j, row in enumerate(data):
            insts.append(
                [
                    row[headers.index("type")],
                    row[headers.index("subtype")],
                    row[headers.index("name")],
                ]
            )

        data = qres["instances"]["data"]
        headers = qres["instances"]["headers"]

        insts = []
        for j, row in enumerate(data):
            insts.append(
                [
                    row[headers.index("type")],
                    row[headers.index("subtype")],
                    row[headers.index("name")],
                ]
            )

        terms = []
        for j, row in enumerate(insts):
            if "term" == row[0]:
                term = row[2].split()
                terms.append(term)

        # qry = andromeda_glm.glm_query()
        qry = glm_query()
        qry.select({"nodes": terms})
        qry.traverse({"edge": "to-doc"})

        config = qry.to_config()
        # print("query: ", json.dumps(config, indent=2))

        out = glm_mdl.query(config)
        # print(json.dumps(out, indent=2))

        if out["status"] == "success":
            docs = pd.DataFrame(
                out["result"][1]["nodes"]["data"],
                columns=out["result"][1]["nodes"]["headers"],
            )

            doc_hashes = list(docs["text"])

            for k, v in topk.items():
                if k == 0:
                    topk[k] += 1
                elif doc_hash in doc_hashes[0:k]:
                    topk[k] += 1

    print(json.dumps(topk, indent=2))

    if show:
        try:
            import matplotlib.pyplot as plt
        except ImportError:
            print(
                "matplotlib is not installed, please install it via python -m pip install matplotlib to activate this functionality."
            )
            sys.exit(1)

        x = []
        y = []
        for i in range(1, 10):
            x.append(i)
            y.append(topk[i] / topk[0])

        plt.figure(1)
        plt.plot(x, y, "r.-", label="doc-topk")
        plt.ylim(0, 1.05)
        plt.legend(loc="lower right")
        plt.show()


def compute_topk_on_element(df, nlp_mdl, glm_mdl):
    """Function to compute topk elements"""

    topk = {0: 0}
    for ind in range(1, 10):
        topk[ind] = 0

    for i, row in df.iterrows():
        print(i, row["question"])

        doc_hash = row["doc_hash"]
        question = row["question"]

        context = row["text"]

        qres = nlp_mdl.apply_nlp_on_text(question)
        cres = nlp_mdl.apply_nlp_on_text(context)

        data = qres["instances"]["data"]
        headers = qres["instances"]["headers"]

        insts = []
        for j, row in enumerate(data):
            insts.append(
                [
                    row[headers.index("type")],
                    row[headers.index("subtype")],
                    row[headers.index("name")],
                ]
            )
        """
        print(f"question: {question}\n")
            
        print("instances: ")
        print(tabulate(insts, headers=["type", "subtype", "name"]), "\n")
        """

        terms = []
        for j, row in enumerate(insts):
            if "term" == row[0]:
                term = row[2].split()
                terms.append(term)

        # qry = andromeda_glm.glm_query()
        qry = glm_query()
        qry.select({"nodes": terms})
        qry.traverse({"edge": "to-text"})

        config = qry.to_config()
        # print("query: ", json.dumps(config, indent=2))

        out = glm_mdl.query(config)
        # print(json.dumps(out, indent=2))

        if out["status"] == "success":
            docs = pd.DataFrame(
                out["result"][1]["nodes"]["data"],
                columns=out["result"][1]["nodes"]["headers"],
            )

            doc_hashes = list(docs["text"])

            for k, v in topk.items():
                if k == 0:
                    topk[k] += 1
                elif doc_hash in doc_hashes[0:k]:
                    topk[k] += 1

    print(json.dumps(topk, indent=2))


if __name__ == "__main__":
    glm_dir, qa_pairs_file, models, show = parse_arguments()

    glm_mdl = load_glm(glm_dir)
    # nlp_mdl = load_nlp(models)
    nlp_mdl = init_nlp_model(
        model_names="name;conn;verb;term;language;reference;abbreviation"
    )

    df = pd.read_csv(qa_pairs_file)

    compute_topk_on_documents(df, nlp_mdl, glm_mdl, show=show)
    # compute_topk_on_element(df, nlp_mdl, glm_mdl)



#!/usr/bin/env python
"""Module to query Deep Search"""

import argparse
import sys

import pandas as pd

from deepsearch_glm.utils.ds_utils import (
    create_docs_dir,
    ds_index_query,
    ds_list_indices,
)


def parse_arguments():
    """Parsing arguments for querying Deep Search indices"""

    parser = argparse.ArgumentParser(
        prog="ds_query",
        description="Query and download documents on the Deep Search platform",
        epilog="""
examples of execution: 

1. search for documents:

    poetry run python ./deepsearch_glm/ds_query.py --index esg-reports --query "\\\"net zero\\\""
    poetry run python ./deepsearch_glm/ds_query.py --index patent-uspto --query "\\\"global warming potential\\\" AND \\\"etching\\\""
    poetry run python ./deepsearch_glm/ds_query.py --index arxiv --query "\\\"quantum computing\\\""


""",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.add_argument(
        "--index", required=True, type=str, help="Deep Search document index"
    )

    parser.add_argument("--query", required=True, type=str, help="Query for document")

    parser.add_argument(
        "--output-dir",
        required=False,
        type=bool,
        default=create_docs_dir(),
        help="output directory for documents",
    )

    args = parser.parse_args()

    return args.index, args.query, args.output_dir


if __name__ == "__main__":
    index, query, odir = parse_arguments()

    indices = ds_list_indices()

    index_found = False
    for item in indices:
        if item["Index"] == index:
            index_found = True

    if not index_found:
        print(pd.DataFrame(indices))
        print(f"index `{index}` not found")
        sys.exit(-1)

    else:
        odir = ds_index_query(index, query, odir, force=True)
        print(f"files downloaded: {odir}")



#!/usr/bin/env python
"""Module to create GLM from PDF/JSON documets"""

import argparse

# import datetime
import glob

# import json
# import os
import sys

from deepsearch_glm.glm_utils import create_glm_dir, create_glm_from_docs
from deepsearch_glm.utils.ds_utils import convert_pdffiles

# import textwrap

# from tabulate import tabulate


def parse_arguments():
    """Function to parse arguments for `create_glm_from_docs`"""

    parser = argparse.ArgumentParser(
        prog="create_glm_from_docs",
        description="Create GLM from Deep Search documents",
        epilog="""
examples of execution: 

1.a run on single document (pdf or json) with default NLP models (=`term`):
    poetry run python ./deepsearch_glm/create_glm_from_docs.py --pdf-docs './data/documents/articles/2305.02334.pdf'
    poetry run python ./deepsearch_glm/create_glm_from_docs.py --json-docs './data/documents/articles/2305.02334.json'

2. run on multiple documents:
    poetry run python ./deepsearch_glm/create_glm_from_docs.py --pdf-docs './data/documents/articles/*.pdf'
    poetry run python ./deepsearch_glm/create_glm_from_docs.py --json-docs './data/documents/articles/*.json'

3. run on multiple documents with non-default models:
    poetry run python ./deepsearch_glm/create_glm_from_docs.py --pdf-docs './data/documents/articles/2305.*.pdf' --models 'language;term;abbreviation'

""",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.add_argument(
        "--pdf-docs",
        required=False,
        type=str,
        default=None,
        help="filename(s) of pdf document",
    )

    parser.add_argument(
        "--json-docs",
        required=False,
        type=str,
        default=None,
        help="filename(s) of json document",
    )

    parser.add_argument(
        "--models",
        required=False,
        type=str,
        default="name;verb;term;abbreviation",
        help="set NLP models (e.g. `term;sentence`)",
    )

    parser.add_argument(
        "--force-convert",
        required=False,
        type=bool,
        default=False,
        help="force pdf conversion",
    )

    parser.add_argument(
        "--output-dir",
        required=False,
        type=str,
        default=create_glm_dir(),
        help="output root directory for GLM",
    )

    args = parser.parse_args()

    pdf_docs = args.pdf_docs
    json_docs = args.json_docs

    if pdf_docs is None and json_docs is None:
        sys.exit(-1)

    pdf_files = []
    if pdf_docs is not None:
        pdf_files = sorted(glob.glob(pdf_docs))

    json_files = []
    if json_docs is not None:
        json_files = sorted(glob.glob(json_docs))

    if len(pdf_files) == 0 and len(json_files) == 0:
        sys.exit(-1)

    return pdf_files, json_files, args.models, args.force_convert, args.output_dir


if __name__ == "__main__":
    pdf_files, json_files, model_names, force_convert, odir = parse_arguments()

    if len(pdf_files) > 0:
        new_json_files = convert_pdffiles(pdf_files, force=force_convert)

        for _ in new_json_files:
            json_files.append(_)

    json_files = sorted(list(set(json_files)))

    glm = create_glm_from_docs(odir, json_files, model_names)
    print(f" --> GLM saved to: {odir}")



#!/usr/bin/env python
"""Module to apply NLP-models on text"""

import argparse

from deepsearch_glm.andromeda_nlp import nlp_model
from deepsearch_glm.nlp_utils import print_on_shell

# import glob
# import json
# import os


DEFAULT_TEXT = """A team of physicists at Université Paris-Saclay has,
for the first time, observed spontaneous quasi-crystal self-assembly.
The observation occurred during an experiment they were conducting with
tiny vibrating magnetic spheres. The team has written a paper describing
their experiment and have posted it on the arXiv preprint server while
they await peer review."""


def parse_arguments():
    """Function to parse arguments for `nlp_apply_on_text`"""

    parser = argparse.ArgumentParser(
        prog="nlp_apply_on_text",
        description="Apply NLP on text",
        epilog="""
examples of execution:

1. apply NLP on default text with default models (=`language`):

    poetry run python ./deepsearch_glm/apply_nlp_on_text.py

2. apply NLP on a single piece of text with default model (=`language`):

    poetry run python ./deepsearch_glm/apply_nlp_on_text.py --text "FeSe is a superconductor with Tc of 30 K."

3. apply NLP on default text with specific models (=`verb;term;conn;semantic;abbreviation`):

    poetry run python ./deepsearch_glm/apply_nlp_on_text.py --model-names "verb;term;conn;semantic;abbreviation"

4. apply NLP on text from prompt with specific models (=`verb;term`):

    poetry run python ./deepsearch_glm/apply_nlp_on_text.py --model-names "verb;term" --interactive True

""",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.add_argument(
        "-t",
        "--text",
        required=False,
        type=str,
        default=DEFAULT_TEXT,
        help="text on which the NLP is applied.",
    )

    parser.add_argument(
        "-m",
        "--model-names",
        required=False,
        type=str,
        default="verb;term;conn;semantic;abbreviation",
        help="model_names to be applied",
    )

    parser.add_argument(
        "--interactive",
        required=False,
        type=bool,
        default=False,
        help="run on text interactively",
    )

    args = parser.parse_args()

    return args.text, args.model_names, args.interactive


if __name__ == "__main__":
    text, model_names, interactive = parse_arguments()

    model = nlp_model()
    # print("resource-dir: ", model.get_resources_path())

    model.initialise_models(model_names)

    if interactive:
        while True:
            text = input("text: ")

            if text in ["q", "quit"]:
                break

            result = model.apply_on_text(text)
            print_on_shell(text, result)

    else:
        result = model.apply_on_text(text)
        print_on_shell(text, result)



#!/usr/bin/env python
"""Module to query the GLM"""

import argparse
import json

import pandas as pd
from tabulate import tabulate

from deepsearch_glm.andromeda_glm import glm_query
from deepsearch_glm.glm_utils import load_glm
from deepsearch_glm.nlp_utils import init_nlp_model


def parse_arguments():
    """Function to parse arguments for `glm_docqa`"""

    parser = argparse.ArgumentParser(
        prog="glm_docqa",
        description="Do Q&A on pdf document",
        epilog="Text at the bottom of help",
    )

    parser.add_argument(
        "--glm-dir", required=True, type=str, help="directory of GLM model"
    )

    args = parser.parse_args()

    return args.glm_dir


def execute_query(glm, query):
    """Exexcute query on GLM"""

    config = query.to_config()
    # print("query: ", json.dumps(config, indent=2))

    out = glm.query(config)
    # print(json.dumps(out, indent=2))

    return out


def display_result(out):
    """Display the GLM-query output"""

    if out["status"] == "success":
        """
        docs = pd.DataFrame(
            out["result"][1]["nodes"]["data"],
            columns=out["result"][1]["nodes"]["headers"],
        )
        print(docs)
        """
        data = out["result"][1]["nodes"]["data"]
        headers = out["result"][1]["nodes"]["headers"]
        print(tabulate(data, headers=headers))


def run(glm):
    """interactively query GLM"""

    while True:
        line = input("prompt: ").strip()
        parts = line.split(" ")

        qry = glm_query()
        # qry.select({"nodes": [["This"]]})
        qry.select({"nodes": [parts]})
        qry.traverse({"edge": "tax-up"})

        out = execute_query(glm, qry)

        display_result(out)

        # break


if __name__ == "__main__":
    glm_dir = parse_arguments()

    glm_mdl = load_glm(glm_dir)

    run(glm_mdl)



#!/usr/bin/env python
"""Module to train CRF reference model"""

import argparse
import glob
import json
import os
import random
import subprocess
import sys
import time

import pandas as pd
import tqdm
from rich import Console
from tabulate import tabulate

# from deepsearch_glm.andromeda_nlp import nlp_model
from deepsearch_glm.nlp_utils import (
    create_nlp_dir,
    get_max_items,
    init_nlp_model,
    train_crf,
)
from deepsearch_glm.utils.load_pretrained_models import get_resources_dir

console = Console()


def parse_arguments():
    """Function to parse arguments for nlp_train_reference"""

    parser = argparse.ArgumentParser(
        prog="nlp_train_reference",
        description="Prepare CRF data for CRF-reference parser",
        epilog="""
examples of execution: 

1. end-to-end example on pdf documents:

    poetry run python ./deepsearch_glm/nlp_train_semantic.py -m all --input-dir '<root-dir-of-json-docs> --output-dir <models-directory>'

2. annotate (100) references:

    poetry run python ./deepsearch_glm/nlp_train_semantic.py -m annotate --input-dir '<root-dir-of-json-docs> --output-dir <models-directory> --max-items 100'        
""",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.add_argument(
        "-m",
        "--mode",
        required=True,
        default="all",
        help="mode for training semantic model",
        choices=["extract", "annotate", "train", "all"],
    )

    parser.add_argument(
        "--input-dir",
        required=True,
        type=str,
        default=None,
        help="input directory with documents",
    )

    parser.add_argument(
        "--output-dir",
        required=False,
        type=str,
        default=None,
        help="output directory for trained models",
    )

    parser.add_argument(
        "--max-items", required=False, type=int, default=-1, help="number of references"
    )

    args = parser.parse_args()

    idir = args.input_dir

    if not os.path.exists(idir):
        print(f"input directory {idir} does not exist")
        sys.exit(-1)

    if args.output_dir is None:
        odir = idir  # create_nlp_dir()

    elif not os.path.exists(args.output_dir):
        os.mkdir(args.output_dir)
        odir = args.output_dir

    else:
        odir = args.output_dir

    return args.mode, idir, odir, args.max_items


def shorten_text(text):
    """Function to shorten text"""

    ntext = text.replace("\n", "")

    return ntext.strip()


def extract_references(filenames, ofile, max_items: int = -1):
    """Function to identify and extract references"""

    nlp_model = init_nlp_model("semantic")

    fw = open(ofile, "w", encoding="utf-8")

    total = 0
    for filename in tqdm.tqdm(filenames):
        # print(f"reading {filename}")

        try:
            with open(filename, "r", encoding="utf-8") as fr:
                idoc = json.load(fr)
        except Exception as exc:
            print(f"could not read line: {str(exc)}")
            continue

        training_sample = random.random() < 0.9

        odoc = nlp_model.apply_on_doc(idoc)

        props = pd.DataFrame(
            odoc["properties"]["data"], columns=odoc["properties"]["headers"]
        )

        props_refs = props[props["label"] == "reference"]
        # print(props_refs)
        refs_hash = list(props_refs["subj_hash"])

        texts = pd.DataFrame.from_records(odoc["texts"])
        # print(texts)

        refs = pd.merge(props_refs, texts, how="inner", on=["subj_hash"])
        # print(refs[refs["confidence"]>0.95][["confidence", "text"]])

        for i, ref in refs.iterrows():
            if ref["confidence"] > 0.95 and len(ref["text"]) > 32:
                item = {"training-sample": training_sample, "text": ref["text"]}
                fw.write(json.dumps(item) + "\n")

                total += 1

        if max_items != -1 and total >= max_items:
            break

    fw.close()


def parse_with_anystyle_api(anystyle, refs):
    """Function to parse references with the anystyle API"""

    time.sleep(1)

    tmpfile = "tmp.json"

    payload = {"input": []}
    for ref in refs:
        payload["input"].append(ref["text"])

    """
    anystyle_token = "9fEhg+39p0J60Bs+WTTwTMcqqTFAUYoyjLlp8nEys4wnfgACn0IoqravX8Exsx/+2q1p4sU7636DR22xUeneLg=="
    anystyle_session = "9GFKMlFoJwbMV6W1Z37YFsG9nbXLqmGicXVzL4r5mn4SqTLcf0revMMFvAjfxcjqR8YBnj2M0fgTWBW12kK1KMFcOgZvZnwQv5lZZ3PQgPP9sait9WgoDR72BHqRpbPe0c1B6%2BNFtYE7aqpugLsTupqBuj%2B%2Fef0tbyd84wC61GkVA9Vtz2nSNC90hDliCre%2BZ2gQUc6runu6yt1M4xa0F8kM4Cxt2pN92XB8hRusqGNfsaCsw5JKdU%2FcDFtdh%2BYDSEBz6DjQFfJq81%2FTI%2F4ulku7mlv73vOC7ew%3D--o%2B2gjgNJqgCjYf4V--3mSN%2FKmNt68WTJsxBh9Bww%3D%3D"

    anystyle_agent = "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/109.0.0.0 Safari/537.36"

    assert anystyle_token == anystyle["token"]
    assert anystyle_session == anystyle["session"]
    assert anystyle_agent == anystyle["agent"]
    """

    token = anystyle["token"]
    session = anystyle["session"]
    agent = anystyle["agent"]

    res = subprocess.call(
        [
            "curl",
            "https://anystyle.io/parse?format=json",
            "-H",
            "authority: anystyle.io",
            "-H",
            "accept: application/json, text/plain, */*",
            "-H",
            "accept-language: en-GB,en-US;q=0.9,en;q=0.8",
            "-H",
            "content-type: application/json;charset=UTF-8",
            "-H",
            f"cookie: _any_style_session={session}",
            "-H",
            "origin: https://anystyle.io",
            "-H",
            "referer: https://anystyle.io/",
            "-H",
            'sec-ch-ua: "Not_A Brand";v="99", "Google Chrome";v="109", "Chromium";v="109"',
            "-H",
            "sec-ch-ua-mobile: ?0",
            "-H",
            'sec-ch-ua-platform: "macOS"',
            "-H",
            "sec-fetch-dest: empty",
            "-H",
            "sec-fetch-mode: cors",
            "-H",
            "sec-fetch-site: same-origin",
            "-H",
            f"user-agent: {agent}",
            "-H",
            f"x-csrf-token: {token}",
            "--data-raw",
            json.dumps(payload),
            "--compressed",
            "-o",
            tmpfile,
            "-s",
        ]
    )

    try:
        with open(tmpfile, "r", encoding="utf-8") as fr:
            tmp = json.load(fr)

        if os.path.exists(tmpfile):
            os.remove(tmpfile)

        return tmp
    except Exception as exc:
        console.print("could not call anystyle API endpoint ...", style="red")
        console.print(f" -> error: {str(exc)}", style="yellow")

    if os.path.exists(tmpfile):
        os.remove(tmpfile)

    return []


def update_references(anystyle, refs, label_map, verbose=False):
    """Function to update references"""

    results = parse_with_anystyle_api(anystyle, refs)

    if len(results) != len(refs):
        return

    for j, item in enumerate(results):
        parts = []
        for row in item:
            parts.append(row[1])

        text = " ".join(parts)
        if text != refs[j]["text"]:
            print("WARNING: mismatch text")
            continue

        beg = 0
        for k, row in enumerate(item):
            charlen = len(row[1].encode("utf-8"))

            item[k].append(beg)
            item[k].append(beg + charlen)

            beg += charlen
            beg += 1

        refs[j]["word_tokens"]["headers"].append("true-label")

        for ri, row_i in enumerate(refs[j]["word_tokens"]["data"]):
            label = "__undef__"
            for rj, row_j in enumerate(item):
                if row_j[2] <= row_i[0] and row_i[1] <= row_j[3]:
                    label = row_j[0]
                    break

            if label in label_map:
                label = label_map[label]
            else:
                ##print(label)
                label = "null"

            refs[j]["word_tokens"]["data"][ri].append(label)

        if verbose:
            print(text)
            print(
                "\n\n",
                tabulate(
                    refs[j]["word_tokens"]["data"],
                    headers=refs[j]["word_tokens"]["headers"],
                ),
            )

        refs[j]["annotated"] = True


def annotate_references(rfile, ofile, max_items):
    """Function to annotate references"""

    label_map = {
        "author": "authors",
        "title": "title",
        "container-title": "conference",
        "journal": "journal",
        "date": "date",
        "volume": "volume",
        "pages": "pages",
        "citation-number": "reference-number",
        "note": "note",
        "url": "url",
        "doi": "doi",
        "isbn": "isbn",
        "publisher": "publisher",
    }

    resources_dir = get_resources_dir()
    with open(f"{resources_dir}/data_nlp.json", "r", encoding="utf-8") as fr:
        configs = json.load(fr)
        anystyle = configs["services"]["anystyle"]

    nlp_model = init_nlp_model("semantic", filters=["properties", "word_tokens"])

    max_items = get_max_items(rfile, max_items)

    refs = []

    fr = open(rfile, "r", encoding="utf-8")
    fw = open(ofile, "w", encoding="utf-8")

    for i in tqdm.tqdm(range(0, max_items)):
        line = fr.readline().strip()
        if line is None or len(line) == 0:
            break

        try:
            item = json.loads(line)
            ref = nlp_model.apply_on_text(item["text"])

            ref["training-sample"] = item["training-sample"]

            refs.append(ref)

        except Exception as exc:
            print(f"Could not process (error: {str(exc)}) for line: {line}")
            continue

        if len(refs) >= 16:
            update_references(anystyle, refs, label_map)

            for ref in refs:
                if "annotated" in ref and ref["annotated"]:
                    fw.write(json.dumps(ref) + "\n")

            refs = []

    fr.close()
    fw.close()

    print(f"writing annotation to {ofile}")


def create_reference_model(mode: str, idir: str, odir: str, max_items: int = -1):
    """Function to create reference model"""

    json_files = glob.glob(os.path.join(idir, "*.json"))
    print("#-docs: ", len(json_files))

    sfile = os.path.join(odir, "nlp-references.data.jsonl")
    afile = os.path.join(odir, "nlp-references.annot.jsonl")

    crf_model_file = os.path.join(odir, "crf_reference")
    crf_metrics_file = crf_model_file + ".metrics.txt"

    if mode in ["extract", "all"]:
        extract_references(json_files, sfile, max_items)

    if mode in ["annotate", "all"]:
        annotate_references(sfile, afile, max_items)

    if mode in ["train", "all"]:
        train_crf("reference", afile, crf_model_file, crf_metrics_file)


if __name__ == "__main__":
    mode, idir, odir, max_items = parse_arguments()

    create_reference_model(mode, idir, odir, max_items)
